# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import triton
import triton.language as tl

from protenix.model.tri_attention.autotune import autotune
from protenix.model.tri_attention.autotune_helpers import _attention_fwd_configs


@autotune(
    configs=_attention_fwd_configs,
    key=["H", "HEAD_DIM", "TAG_N"],
)
@triton.jit
def _attention_fwd(
    Q,
    K,
    V,
    O,
    Bias1,
    Bias2,
    M,
    N,
    S,
    H,
    HEAD_DIM: tl.constexpr,
    TAG_N: tl.constexpr,
    deterministic: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    """
    B: batch
    N: number
    S: sequence, equal to N
    H: number of head
    HEAD_DIM: dimension of head

    Q: (B, N, S, H, HEAD_DIM)
    K: (B, N, S, H, HEAD_DIM)
    V: (B, N, S, H, HEAD_DIM)
    O: (B, N, S, H, HEAD_DIM)
    Bias1: (B, N, 1, 1, S), Bias1 is mask which doesn't need grad
    Bias2: (B, 1, H, S, S)

    M: (B, N, S, H)
    """
    input_dtype = Q.dtype.element_ty

    bn = tl.program_id(2)
    h = tl.program_id(1)
    s = tl.program_id(0) * BLOCK_M
    b = bn // N
    n = bn % N

    Bs = N * S * H * HEAD_DIM
    Ns = S * H * HEAD_DIM
    Ss = H * HEAD_DIM
    Hs = HEAD_DIM

    qkvo_offset = b * Bs + n * Ns + h * Hs
    bias1_offset = b * N * S + n * S
    bias2_offset = b * S * S * H + h * S * S
    m_offset = b * N * H * S + n * S * H + h

    q_ptr = tl.make_block_ptr(
        base=Q + qkvo_offset,
        shape=(S, HEAD_DIM),
        strides=(Ss, 1),
        offsets=(s, 0),
        block_shape=(BLOCK_M, HEAD_DIM),
        order=(1, 0),
    )

    k_ptr = tl.make_block_ptr(
        base=K + qkvo_offset,
        shape=(S, HEAD_DIM),
        strides=(Ss, 1),
        offsets=(0, 0),
        block_shape=(BLOCK_N, HEAD_DIM),
        order=(0, 1),
    )

    v_ptr = tl.make_block_ptr(
        base=V + qkvo_offset,
        shape=(S, HEAD_DIM),
        strides=(Ss, 1),
        offsets=(0, 0),
        block_shape=(BLOCK_N, HEAD_DIM),
        order=(1, 0),
    )

    o_ptr = tl.make_block_ptr(
        base=O + qkvo_offset,
        shape=(S, HEAD_DIM),
        strides=(Ss, 1),
        offsets=(s, 0),
        block_shape=(BLOCK_M, HEAD_DIM),
        order=(1, 0),
    )

    bias1_ptr = tl.make_block_ptr(
        base=Bias1 + bias1_offset,
        shape=(S, S),
        strides=(0, 1),
        offsets=(s, 0),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )

    bias2_ptr = tl.make_block_ptr(
        base=Bias2 + bias2_offset,
        shape=(S, S),
        strides=(S, 1),
        offsets=(s, 0),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )

    m_ptr = tl.make_block_ptr(
        base=M + m_offset,
        shape=(S, 1),
        strides=(H, 1),
        offsets=(s, 0),
        block_shape=(BLOCK_M, 1),
        order=(0, 1),
    )

    qk_scale = 1.4426950408889634

    m_i = tl.zeros([BLOCK_M], dtype=tl.float32) - float("inf")
    l_i = tl.zeros([BLOCK_M], dtype=tl.float32) + 1
    acc = tl.zeros([BLOCK_M, HEAD_DIM], dtype=tl.float32)

    q = tl.load(q_ptr, boundary_check=(0, 1), padding_option="zero")

    offset_q = (s + tl.arange(0, BLOCK_M))[:, None] < S
    offset_k = tl.arange(0, BLOCK_N)[None, :]
    for _ in range(tl.cdiv(S, BLOCK_N)):
        k = tl.load(k_ptr, boundary_check=(0, 1), padding_option="zero")
        bias1 = tl.load(bias1_ptr, boundary_check=(0, 1), padding_option="zero")
        bias2 = tl.load(bias2_ptr, boundary_check=(0, 1), padding_option="zero")
        mask = offset_q * (offset_k < S)
        qk = bias1.to(tl.float32) + bias2.to(tl.float32)
        qk = (
            tl.dot(
                q, tl.trans(k), qk, input_precision="ieee" if deterministic else "tf32"
            )
            * qk_scale
        )
        qk = tl.where(mask, qk, -1e6)
        m_ij = tl.maximum(tl.max(qk, 1), m_i)
        qk = qk - m_ij[:, None]
        alpha = tl.math.exp2(m_i - m_ij)
        p = tl.math.exp2(qk)
        l_ij = l_i * alpha + tl.sum(p, 1)
        acc = acc * alpha[:, None]

        v = tl.load(v_ptr, boundary_check=(0, 1), padding_option="zero")
        p = p.to(input_dtype)
        acc = tl.dot(p, v, acc, input_precision="ieee" if deterministic else "tf32")

        l_i = l_ij
        m_i = m_ij

        k_ptr = tl.advance(k_ptr, (BLOCK_N, 0))
        v_ptr = tl.advance(v_ptr, (BLOCK_N, 0))
        bias1_ptr = tl.advance(bias1_ptr, (0, BLOCK_N))
        bias2_ptr = tl.advance(bias2_ptr, (0, BLOCK_N))
        offset_k += BLOCK_N

    acc = acc / l_i[:, None]
    m_i += tl.math.log2(l_i)
    acc = acc.to(input_dtype)
    tl.store(o_ptr, acc, boundary_check=(0, 1))
    tl.store(m_ptr, m_i[:, None], boundary_check=(0, 1))



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os

import pandas as pd
import torch
from esm import FastaBatchedDataset, pretrained
from tqdm.auto import tqdm

ESM_CONFIG = {
    "esm2-3b": {
        "type": "esm2",
        "model_path": "esm2_t36_3B_UR50D.pt",
        "emb_dim": 2560,
        "n_layers": 36,
    },
    "esm2-3b-ism": {
        "type": "esm2",
        "model_path": "esm2_t36_3B_UR50D_ism.pt",
        "emb_dim": 2560,
        "n_layers": 36,
    },  # https://www.biorxiv.org/content/10.1101/2024.11.08.622579v2
}


def _load_esm2_model(model_path):
    if os.path.exists(model_path):
        model, alphabet = pretrained.load_model_and_alphabet_local(model_path)
    else:
        model, alphabet = pretrained.load_model_and_alphabet(
            os.path.splitext(os.path.basename(model_path))[0]
        )
    return model, alphabet


def load_esm_model(model_name, local_esm_dir="release_data/checkpoint"):
    local_model_path = os.path.join(local_esm_dir, ESM_CONFIG[model_name]["model_path"])
    if os.path.exists(local_model_path):
        print("Try to load ESM language model from ", local_model_path)

    if "ism" in model_name and not os.path.exists(local_model_path):
        raise RuntimeError(
            f"esm2-3b-ism model: {local_model_path} does not exist \n"
            + "this model can not be download from fair-esm, \n"
            + "download it from https://af3-dev.tos-cn-beijing.volces.com/release_model/esm2_t36_3B_UR50D_ism.pt"
        )
    if model_name.startswith("esm2"):
        model, alphabet = _load_esm2_model(local_model_path)
    model.eval()
    if torch.cuda.is_available():
        model = model.cuda()

    return model, alphabet


def _check_files_exist(save_dir, labels):
    return all(
        [os.path.exists(os.path.join(save_dir, label + ".pt")) for label in labels]
    )


def compute_ESM_embeddings(
    model_name,
    model,
    alphabet,
    labels,
    sequences,
    save_dir,
    toks_per_batch=4096,
    truncation_seq_length=1022,
):
    if model_name.startswith("esm2"):
        embeddings = compute_esm2_embeddings(
            model,
            alphabet,
            labels,
            sequences,
            save_dir,
            toks_per_batch,
            truncation_seq_length,
        )
    return embeddings


# Adapt from Corso, Gabriele, et al. "Diffdock: Diffusion steps, twists, and turns for molecular docking."
# URL: https://github.com/gcorso/DiffDock/blob/main/utils/inference_utils.py
def compute_esm2_embeddings(
    model,
    alphabet,
    labels,
    sequences,
    save_dir,
    toks_per_batch=4096,
    truncation_seq_length=1022,
):
    dataset = FastaBatchedDataset(labels, sequences)
    batches = dataset.get_batch_indices(toks_per_batch, extra_toks_per_seq=1)
    data_loader = torch.utils.data.DataLoader(
        dataset,
        collate_fn=alphabet.get_batch_converter(truncation_seq_length),
        batch_sampler=batches,
    )
    repr_layer = model.num_layers
    embeddings = {}
    with torch.no_grad():
        for batch_idx, (labels, strs, toks) in enumerate(tqdm(data_loader)):
            print(
                f"Processing {batch_idx + 1} of {len(batches)} batches ({toks.size(0)} sequences)"
            )
            if _check_files_exist(save_dir, labels):
                continue
            if torch.cuda.is_available():
                toks = toks.to(device="cuda", non_blocking=True)
            out = model(toks, repr_layers=[repr_layer], return_contacts=False)
            representation = out["representations"][repr_layer].to(device="cpu")
            for i, label in enumerate(labels):
                truncate_len = min(truncation_seq_length, len(strs[i]))
                embeddings[label] = representation[i, 1 : truncate_len + 1].clone()
                save_path = os.path.join(save_dir, label + ".pt")
                torch.save(embeddings[label], save_path)
    return embeddings


def pdb_sequences_iterator(
    input_path="./scripts/msa/data/pdb_seqs/pdb_seq.csv",
    save_path="./scripts/msa/data/pdb_seqs/pdb_labels_seqs.csv",
    start_id=0,
    end_id=-1,
):
    if os.path.exists(save_path):
        df_seq = pd.read_csv(save_path)
    else:
        df = pd.read_csv(input_path)
        # Protein only
        df = df[df["mol_type"] == "protein"]
        # Sequence name
        df["pdb_entity_id"] = df["pdb_id"] + "_" + df["entity_id"].astype(str)
        # Group by 'seq'
        df_seq = (
            df.groupby("seq")["pdb_entity_id"]
            .apply(lambda x: ",".join(x))
            .reset_index()
        )
        # Use the first pdb_entity_id as the label
        df_seq["seq_label"] = df_seq["pdb_entity_id"].apply(lambda x: x.split(",")[0])
        assert df_seq["seq_label"].nunique() == len(df_seq)
        # Get a part id
        df_seq["part_id"] = df_seq["pdb_entity_id"].apply(lambda x: x[1:3])
        df_seq.to_csv(save_path)

    if end_id == -1:
        end_id = len(df_seq)
    df_seq = df_seq[start_id:end_id]

    part_counts = dict(df_seq["part_id"].value_counts())
    for part_id, count in part_counts.items():
        df_part = df_seq[df_seq["part_id"] == part_id]
        print(f"Part {part_id}: {len(df_part)} sequences.")
        yield part_id, df_part["seq_label"].tolist(), df_part["seq"].tolist()


def process_pdb_dataset(
    model_name, root_save_dir, pdb_seq_path, pdb_seq_label_path, start_id=0, end_id=-1
):

    model, alphabet = load_esm_model(model_name)
    seq_iterator = pdb_sequences_iterator(
        pdb_seq_path, pdb_seq_label_path, start_id, end_id
    )
    error_parts = []
    for part_id, labels, sequences in seq_iterator:
        save_dir = os.path.join(root_save_dir, f"{part_id}")

        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        print(f"[{part_id}] Generating ESM language model embeddings")
        lm_embeddings = compute_ESM_embeddings(
            model_name,
            model,
            alphabet,
            labels,
            sequences,
            save_dir,
            truncation_seq_length=4094,
            toks_per_batch=16384,
        )
        print(f"[{part_id}] Processed {len(lm_embeddings)} sequences in total. Done!")

    print("Error parts: ", error_parts)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--model_name", type=str, choices=list(ESM_CONFIG.keys()))
    parser.add_argument("--start_id", type=int, default=0)
    parser.add_argument("--end_id", type=int, default=-1)
    args = parser.parse_args()

    save_dir = f"./esm_embeddings/{args.model_name}"
    pdb_seq_path = "./scripts/msa/data/pdb_seqs/pdb_seq.csv"
    pdb_seq_label_path = "./scripts/msa/data/pdb_seqs/pdb_labels_seqs.csv"

    if not os.path.exists(save_dir):
        print("Make dir: ", save_dir)
        os.makedirs(save_dir)
    process_pdb_dataset(
        args.model_name,
        save_dir,
        pdb_seq_path,
        pdb_seq_label_path,
        args.start_id,
        args.end_id,
    )


if __name__ == "__main__":
    main()



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import multiprocessing
import os
import pathlib
import shutil
import time
from concurrent import futures
from pathlib import Path
from typing import Any, Optional, Union

from protenix.data.constants import RNA_CHAIN
from protenix.data.msa.msa_utils import RawMsa
from protenix.data.tools.search import (
    DatabaseConfig,
    Jackhmmer,
    JackhmmerConfig,
    MsaTool,
    Nhmmer,
    NhmmerConfig,
    RunConfig,
)
from protenix.utils.logger import get_logger

logger = get_logger(__name__)

RFAM_SEARCH_DATABASE_URL = "https://protenix.tos-cn-beijing.volces.com/search_database/rfam_14_9_clust_seq_id_90_cov_80_rep_seq.fasta"
RNACENTRAL_SEARCH_DATABASE_URL = "https://protenix.tos-cn-beijing.volces.com/search_database/rnacentral_active_seq_id_90_cov_80_linclust.fasta"
NT_SEARCH_DATABASE_URL = "https://protenix.tos-cn-beijing.volces.com/search_database/nt_rna_2023_02_23_clust_seq_id_90_cov_80_rep_seq.fasta"


def get_msa_tool(
    msa_tool_config: Union[JackhmmerConfig, NhmmerConfig],
) -> MsaTool:
    """
    Returns the requested MSA tool based on the provided configuration.

    Args:
        msa_tool_config: Configuration for Jackhmmer or Nhmmer.

    Returns:
        An instance of MsaTool (Jackhmmer or Nhmmer).

    Raises:
        ValueError: If the configuration type is unknown.
    """
    if isinstance(msa_tool_config, JackhmmerConfig):
        return Jackhmmer(
            binary_path=msa_tool_config.binary_path,
            database_path=msa_tool_config.database_config.path,
            n_cpu=msa_tool_config.n_cpu,
            n_iter=msa_tool_config.n_iter,
            e_value=msa_tool_config.e_value,
            z_value=msa_tool_config.z_value,
            max_sequences=msa_tool_config.max_sequences,
        )
    elif isinstance(msa_tool_config, NhmmerConfig):
        return Nhmmer(
            binary_path=msa_tool_config.binary_path,
            hmmalign_binary_path=msa_tool_config.hmmalign_binary_path,
            hmmbuild_binary_path=msa_tool_config.hmmbuild_binary_path,
            database_path=msa_tool_config.database_config.path,
            n_cpu=msa_tool_config.n_cpu,
            e_value=msa_tool_config.e_value,
            max_sequences=msa_tool_config.max_sequences,
            alphabet=msa_tool_config.alphabet,
        )
    else:
        raise ValueError(f"Unknown MSA tool: {msa_tool_config}.")


def get_msa(
    target_sequence: str,
    run_config: RunConfig,
    chain_poly_type: str,
    deduplicate: bool = False,
) -> RawMsa:
    """
    Computes the MSA for a given query sequence.

    Args:
        target_sequence: The target amino-acid or nucleotide sequence.
        run_config: MSA run configuration.
        chain_poly_type: The type of chain for which to get an MSA.
        deduplicate: If True, the MSA sequences will be deduplicated in the input
            order. Lowercase letters (insertions) are ignored when deduplicating.

    Returns:
        Aligned RawMsa object.
    """
    return RawMsa.from_a3m(
        query=target_sequence,
        ctype=chain_poly_type,
        a3m=get_msa_tool(run_config.config).query(target_sequence).a3m,
        depth_limit=run_config.crop_size,
        dedup=deduplicate,
    )


# Cache to avoid re-running the Nhmmer for the same sequence in homomers.
@functools.cache
def _get_rna_msa(
    sequence: str,
    nt_rna_msa_config: RunConfig,
    rfam_msa_config: RunConfig,
    rnacentral_msa_config: RunConfig,
) -> RawMsa:
    """
    Processes a single RNA chain by running multiple MSA tools in parallel.

    Args:
        sequence: The RNA sequence.
        nt_rna_msa_config: Config for NT-RNA database search.
        rfam_msa_config: Config for Rfam database search.
        rnacentral_msa_config: Config for RNAcentral database search.

    Returns:
        Merged RawMsa object.
    """
    logger.info("Getting RNA MSAs for sequence %s", sequence)
    rna_msa_start_time = time.time()

    # Run various MSA tools in parallel. Use a ThreadPoolExecutor because
    # they're not blocked by the GIL, as they're sub-shelled out.
    with futures.ThreadPoolExecutor() as executor:
        # Currently only Rfam search is enabled in the provided logic
        rfam_msa_future = executor.submit(
            get_msa,
            target_sequence=sequence,
            run_config=rfam_msa_config,
            chain_poly_type=RNA_CHAIN,
        )
        rna_central_msa_future = executor.submit(
            get_msa,
            target_sequence=sequence,
            run_config=rnacentral_msa_config,
            chain_poly_type=RNA_CHAIN,
        )
        nt_rna_msa_future = executor.submit(
            get_msa,
            target_sequence=sequence,
            run_config=nt_rna_msa_config,
            chain_poly_type=RNA_CHAIN,
        )

    rfam_msa = rfam_msa_future.result()
    rna_central_msa = rna_central_msa_future.result()
    nt_rna_msa = nt_rna_msa_future.result()

    rna_msa = RawMsa.merge(
        msas=[rfam_msa, rna_central_msa, nt_rna_msa],
        deduplicate=True,
    )

    logger.info(
        "Getting RNA MSAs took %.2f seconds for sequence %s, found %d unpaired"
        " sequences",
        time.time() - rna_msa_start_time,
        sequence,
        rna_msa.depth,
    )
    return rna_msa


def run_rna_msa_search(
    rna_seq_for_msa_search: Optional[str] = None,
    rna_result_path: Optional[str] = None,
    rna_seq_id: str = "example",
    nhmmer_binary_path: Optional[str] = None,
    hmmalign_binary_path: Optional[str] = None,
    hmmbuild_binary_path: Optional[str] = None,
    ntrna_database_path: Optional[str] = None,
    rfam_database_path: Optional[str] = None,
    rna_central_database_path: Optional[str] = None,
    nhmmer_n_cpu: Optional[int] = None,
) -> None:
    """
    Run RNA MSA search using nhmmer against multiple databases.

    Args:
        rna_seq_for_msa_search (Optional[str]): RNA sequence to search.
        rna_result_path (Optional[str]): Output path for MSA results.
        rna_seq_id (str): Identifier for the RNA sequence.
        nhmmer_binary_path (Optional[str]): Path to nhmmer binary.
        hmmalign_binary_path (Optional[str]): Path to hmmalign binary.
        hmmbuild_binary_path (Optional[str]): Path to hmmbuild binary.
        ntrna_database_path (Optional[str]): NT-RNA database path.
        rfam_database_path (Optional[str]): Rfam database path.
        rna_central_database_path (Optional[str]): RNAcentral database path.
        nhmmer_n_cpu (Optional[int]): Number of CPUs for nhmmer.
    """
    assert rna_result_path is not None, "output path should not be None"
    assert rna_seq_for_msa_search is not None, "RNA sequence should not be None"

    if nhmmer_binary_path is None:
        nhmmer_binary_path = shutil.which("nhmmer")
        if nhmmer_binary_path is None:
            raise AssertionError(
                "nhmmer binary path should not be None. You can install "
                "hmmer using: apt install hmmer or conda install -c bioconda hmmer"
            )
    else:
        if not os.path.exists(nhmmer_binary_path):
            raise AssertionError(
                f"nhmmer binary path {nhmmer_binary_path} does not exist"
            )

    if hmmalign_binary_path is None:
        hmmalign_binary_path = shutil.which("hmmalign")
        if hmmalign_binary_path is None:
            raise AssertionError(
                "hmmalign binary path should not be None. You can install "
                "hmmer using: apt install hmmer or conda install -c bioconda hmmer"
            )
    else:
        if not os.path.exists(hmmalign_binary_path):
            raise AssertionError(
                f"hmmalign binary path {hmmalign_binary_path} does not exist"
            )

    if hmmbuild_binary_path is None:
        hmmbuild_binary_path = shutil.which("hmmbuild")
        if hmmbuild_binary_path is None:
            raise AssertionError(
                "hmmbuild binary path should not be None. You can install "
                "hmmer using: apt install hmmer or conda install -c bioconda hmmer"
            )
    else:
        if not os.path.exists(hmmbuild_binary_path):
            raise AssertionError(
                f"hmmbuild binary path {hmmbuild_binary_path} does not exist"
            )

    if nhmmer_n_cpu is None:
        nhmmer_n_cpu = min(multiprocessing.cpu_count(), 8)

    if ntrna_database_path is None:
        _HOME_DIR = pathlib.Path(os.environ.get("PROTENIX_ROOT_DIR", str(Path.home())))
        ntrna_database_path = (
            _HOME_DIR
            / "search_database"
            / "nt_rna_2023_02_23_clust_seq_id_90_cov_80_rep_seq.fasta"
        ).as_posix()
    if not os.path.exists(ntrna_database_path):
        from runner.inference import download_from_url

        os.makedirs(os.path.dirname(ntrna_database_path), exist_ok=True)
        logger.info(
            f"Downloading nt-rna database from {NT_SEARCH_DATABASE_URL} to {ntrna_database_path}"
        )
        download_from_url(
            NT_SEARCH_DATABASE_URL, ntrna_database_path, check_weight=False
        )

    if rfam_database_path is None:
        _HOME_DIR = pathlib.Path(os.environ.get("PROTENIX_ROOT_DIR", str(Path.home())))
        rfam_database_path = (
            _HOME_DIR
            / "search_database"
            / "rfam_14_9_clust_seq_id_90_cov_80_rep_seq.fasta"
        ).as_posix()
    if not os.path.exists(rfam_database_path):
        from runner.inference import download_from_url

        os.makedirs(os.path.dirname(rfam_database_path), exist_ok=True)
        logger.info(
            f"Downloading rfam database from {RFAM_SEARCH_DATABASE_URL} to {rfam_database_path}"
        )
        download_from_url(
            RFAM_SEARCH_DATABASE_URL, rfam_database_path, check_weight=False
        )

    if rna_central_database_path is None:
        _HOME_DIR = pathlib.Path(os.environ.get("PROTENIX_ROOT_DIR", str(Path.home())))
        rna_central_database_path = (
            _HOME_DIR
            / "search_database"
            / "rnacentral_active_seq_id_90_cov_80_linclust.fasta"
        ).as_posix()
    if not os.path.exists(rna_central_database_path):
        from runner.inference import download_from_url

        os.makedirs(os.path.dirname(rna_central_database_path), exist_ok=True)
        logger.info(
            f"Downloading rna-central database from {RNACENTRAL_SEARCH_DATABASE_URL} to {rna_central_database_path}"
        )
        download_from_url(
            RNACENTRAL_SEARCH_DATABASE_URL,
            rna_central_database_path,
            check_weight=False,
        )

    logger.info("RNA MSA search start!")
    rna_msa_start_time = time.time()

    nt_rna_msa_config = RunConfig(
        config=NhmmerConfig(
            binary_path=nhmmer_binary_path,
            hmmalign_binary_path=hmmalign_binary_path,
            hmmbuild_binary_path=hmmbuild_binary_path,
            database_config=DatabaseConfig(
                name="nt_rna",
                path=ntrna_database_path,
            ),
            n_cpu=nhmmer_n_cpu,
            e_value=1e-3,
            alphabet="rna",
            max_sequences=10_000,
        ),
        chain_poly_type=RNA_CHAIN,
        crop_size=None,
    )

    rfam_msa_config = RunConfig(
        config=NhmmerConfig(
            binary_path=nhmmer_binary_path,
            hmmalign_binary_path=hmmalign_binary_path,
            hmmbuild_binary_path=hmmbuild_binary_path,
            database_config=DatabaseConfig(
                name="rfam_rna",
                path=rfam_database_path,
            ),
            n_cpu=nhmmer_n_cpu,
            e_value=1e-3,
            alphabet="rna",
            max_sequences=10_000,
        ),
        chain_poly_type=RNA_CHAIN,
        crop_size=None,
    )

    rnacentral_msa_config = RunConfig(
        config=NhmmerConfig(
            binary_path=nhmmer_binary_path,
            hmmalign_binary_path=hmmalign_binary_path,
            hmmbuild_binary_path=hmmbuild_binary_path,
            database_config=DatabaseConfig(
                name="rna_central_rna",
                path=rna_central_database_path,
            ),
            n_cpu=nhmmer_n_cpu,
            e_value=1e-3,
            alphabet="rna",
            max_sequences=10_000,
        ),
        chain_poly_type=RNA_CHAIN,
        crop_size=None,
    )

    rna_msa = _get_rna_msa(
        sequence=rna_seq_for_msa_search,
        nt_rna_msa_config=nt_rna_msa_config,
        rfam_msa_config=rfam_msa_config,
        rnacentral_msa_config=rnacentral_msa_config,
    )
    rna_msa_a3m = rna_msa.to_a3m()
    os.makedirs(f"{rna_result_path}/{rna_seq_id}", exist_ok=True)
    with open(f"{rna_result_path}/{rna_seq_id}/rna_msa.a3m", "w") as f:
        f.write(rna_msa_a3m)
    rna_msa_end_time = time.time()
    logger.info(
        f"RNA MSA search done!, using {(rna_msa_end_time - rna_msa_start_time):.2f}s"
    )


def update_rna_msa_info(
    json_data: list[dict[str, Any]],
    out_dir: str,
    nhmmer_binary_path: Optional[str] = None,
    hmmalign_binary_path: Optional[str] = None,
    hmmbuild_binary_path: Optional[str] = None,
    ntrna_database_path: Optional[str] = None,
    rfam_database_path: Optional[str] = None,
    rna_central_database_path: Optional[str] = None,
    nhmmer_n_cpu: Optional[int] = None,
) -> bool:
    """
    Update RNA MSA information in the JSON data.
    If unpairedMsaPath is missing, it performs an RNA MSA search.

    Args:
        json_data (list[dict[str, Any]]): The input JSON data.
        out_dir (str): Directory to save RNA MSA results.
        nhmmer_binary_path (Optional[str]): Path to nhmmer binary.
        hmmalign_binary_path (Optional[str]): Path to hmmalign binary.
        hmmbuild_binary_path (Optional[str]): Path to hmmbuild binary.
        ntrna_database_path (Optional[str]): NT-RNA database path.
        rfam_database_path (Optional[str]): Rfam database path.
        rna_central_database_path (Optional[str]): RNAcentral database path.
        nhmmer_n_cpu (Optional[int]): Number of CPUs for nhmmer.

    Returns:
        bool: True if any RNA MSA information was updated.
    """
    actual_updated = False
    for task_idx, infer_data in enumerate(json_data):
        task_name = infer_data.get("name", f"task_{task_idx}")
        for sequence_idx, sequence in enumerate(infer_data["sequences"]):
            if "rnaSequence" in sequence:
                rna_chain = sequence["rnaSequence"]
                rna_sequence = rna_chain.get("sequence", "")
                if not rna_sequence:
                    continue

                if "unpairedMsaPath" in rna_chain and os.path.exists(
                    rna_chain["unpairedMsaPath"]
                ):
                    continue

                logger.info(
                    f"Running RNA MSA search for task {task_name}, rna_sequence: {rna_sequence}"
                )
                rna_output_dir = os.path.join(
                    out_dir, task_name, "rna_msa", str(sequence_idx)
                )
                run_rna_msa_search(
                    rna_seq_for_msa_search=rna_sequence,
                    rna_result_path=os.path.dirname(rna_output_dir),
                    rna_seq_id=os.path.basename(rna_output_dir),
                    nhmmer_binary_path=nhmmer_binary_path,
                    hmmalign_binary_path=hmmalign_binary_path,
                    hmmbuild_binary_path=hmmbuild_binary_path,
                    ntrna_database_path=ntrna_database_path,
                    rfam_database_path=rfam_database_path,
                    rna_central_database_path=rna_central_database_path,
                    nhmmer_n_cpu=nhmmer_n_cpu,
                )
                rna_msa_path = os.path.join(rna_output_dir, "rna_msa.a3m")
                if os.path.exists(rna_msa_path):
                    rna_chain["unpairedMsaPath"] = rna_msa_path
                    actual_updated = True
    return actual_updated


if __name__ == "__main__":
    example_seq = (
        "GGCGCGUUAACAAAGCGGUUAUGUAGCGGAUUGCAAAUCCGUCUAGUCCGGUUCGACUCCGGAACGCGCCUCCA"
    )
    run_rna_msa_search(
        rna_seq_for_msa_search=example_seq,
        rna_result_path="./output",
        rna_seq_id="rna_seq_1",
    )



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2024 ByteDance and/or its affiliates.
#
# Copyright 2021- HPC-AI Technology Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib
import numbers
import os
import sys
from typing import Any, Optional, Union

import torch
from torch.nn.parameter import Parameter

sys.path.append(os.path.dirname(__file__))

try:
    fast_layer_norm_cuda_v2 = importlib.import_module("fast_layer_norm_cuda_v2")
except ImportError:
    from protenix.model.layer_norm.torch_ext_compile import compile

    current_dir = os.path.dirname(__file__)
    fast_layer_norm_cuda_v2 = compile(
        name="fast_layer_norm_cuda_v2",
        sources=[
            os.path.join(f"{current_dir}/kernel", file)
            for file in ["layer_norm_cuda.cpp", "layer_norm_cuda_kernel.cu"]
        ],
        extra_include_paths=[f"{current_dir}/kernel"],
        build_directory=current_dir,
    )


class FusedLayerNormAffineFunction(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx: Any,
        input: torch.Tensor,
        weight: Optional[torch.Tensor],
        bias: Optional[torch.Tensor],
        normalized_shape: torch.Size,
        eps: float,
    ) -> torch.Tensor:
        d = input.dtype

        ctx.normalized_shape = normalized_shape
        ctx.eps = eps
        input_ = input.contiguous()

        if weight is None:
            if bias is None:
                output, mean, invvar = fast_layer_norm_cuda_v2.forward_none_affine(
                    input_, ctx.normalized_shape, ctx.eps
                )
            else:
                output, mean, invvar = fast_layer_norm_cuda_v2.forward_with_bias_affine(
                    input_, ctx.normalized_shape, bias.to(d), ctx.eps
                )
        else:
            if bias is None:
                (
                    output,
                    mean,
                    invvar,
                ) = fast_layer_norm_cuda_v2.forward_with_weight_affine(
                    input_, ctx.normalized_shape, weight.to(d), ctx.eps
                )
            else:
                output, mean, invvar = fast_layer_norm_cuda_v2.forward_with_both_affine(
                    input_,
                    ctx.normalized_shape,
                    weight.to(d),
                    bias.to(d),
                    ctx.eps,
                )
        ctx.save_for_backward(input_, weight, bias, mean, invvar)
        return output

    @staticmethod
    def backward(
        ctx: Any, grad_output: torch.Tensor
    ) -> tuple[Optional[torch.Tensor], ...]:
        d = grad_output.dtype
        input_, weight_, bias_, mean, invvar = ctx.saved_tensors
        grad_input = grad_weight = grad_bias = None

        if weight_ is None:
            if bias_ is None:
                (
                    grad_input,
                    grad_weight,
                    grad_bias,
                ) = fast_layer_norm_cuda_v2.backward_none_affine(
                    grad_output.contiguous(),
                    mean,
                    invvar,
                    input_,
                    ctx.normalized_shape,
                    ctx.eps,
                )
            else:
                (
                    grad_input,
                    grad_weight,
                    grad_bias,
                ) = fast_layer_norm_cuda_v2.backward_with_bias_affine(
                    grad_output.contiguous(),
                    mean,
                    invvar,
                    input_,
                    ctx.normalized_shape,
                    bias_.to(dtype=d),
                    ctx.eps,
                )
        else:
            if bias_ is None:
                (
                    grad_input,
                    grad_weight,
                    grad_bias,
                ) = fast_layer_norm_cuda_v2.backward_with_weight_affine(
                    grad_output.contiguous(),
                    mean,
                    invvar,
                    input_,
                    ctx.normalized_shape,
                    weight_.to(dtype=d),
                    ctx.eps,
                )
            else:
                (
                    grad_input,
                    grad_weight,
                    grad_bias,
                ) = fast_layer_norm_cuda_v2.backward_with_both_affine(
                    grad_output.contiguous(),
                    mean,
                    invvar,
                    input_,
                    ctx.normalized_shape,
                    weight_.to(dtype=d),
                    bias_.to(dtype=d),
                    ctx.eps,
                )
        return (
            grad_input,
            None if weight_ is None else grad_weight,
            None if bias_ is None else grad_bias,
            None,
            None,
            None,
        )


class FusedLayerNorm(torch.nn.Module):
    """
    Args:
        normalized_shape (int or list or torch.Size) input shape from an expected input of size
        create_scale (bool) If set to False, the layer will not learn an additive weight, Default: True
        create_offset (bool) If set to False, the layer will not learn an additive bias, Default: True
        eps (float) a value added to the denominator for numerical stability. Default: 1e-5
    """

    def __init__(
        self,
        normalized_shape: Union[int, list[int], torch.Size],
        create_scale: bool = True,
        create_offset: bool = True,
        eps: float = 1e-5,
    ) -> None:
        super(FusedLayerNorm, self).__init__()

        if isinstance(normalized_shape, numbers.Integral):
            normalized_shape = (normalized_shape,)
        self.normalized_shape = torch.Size(normalized_shape)
        self.eps = eps
        if create_scale:
            self.weight = Parameter(torch.ones(*normalized_shape))
        else:
            self.weight = None

        if create_offset:
            self.bias = Parameter(torch.zeros(*normalized_shape))
        else:
            self.bias = None

        self.reset_parameters()

    def reset_parameters(self) -> None:
        if self.weight is not None:
            torch.nn.init.ones_(self.weight)
        if self.bias is not None:
            torch.nn.init.zeros_(self.bias)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return FusedLayerNormAffineFunction.apply(
            input, self.weight, self.bias, self.normalized_shape, self.eps
        )


if __name__ == "__main__":
    dtype = torch.float32
    data = torch.rand(10, 10).cuda().to(dtype=dtype)
    data1 = data * 1
    data.requires_grad = True
    data1.requires_grad = True
    layer_norm = (
        FusedLayerNorm(10, create_scale=True, create_offset=True).cuda().to(dtype=dtype)
    )
    layer_norm_torch = torch.nn.LayerNorm(10).cuda().to(dtype=dtype)
    out = layer_norm(data)
    out1 = layer_norm_torch(data1)
    # print(out - out1)
    loss = out.sum()
    loss.backward()
    loss1 = out1.sum()
    loss1.backward()
    print(data.grad - data1.grad)
    print(layer_norm.weight.grad - layer_norm_torch.weight.grad)
    print(layer_norm.bias.grad - layer_norm_torch.bias.grad)
    print(layer_norm.weight.grad, layer_norm.bias.grad)

    # layer_norm = FusedLayerNorm(10, create_scale=True, create_offset=False).cuda()
    # out = layer_norm(data)
    # loss = out.sum()
    # loss.backward()

    # layer_norm = FusedLayerNorm(10, create_scale=False, create_offset=True).cuda()
    # out = layer_norm(data)
    # loss = out.sum()
    # loss.backward()

    # layer_norm = FusedLayerNorm(10, create_scale=True, create_offset=True).cuda()
    # out = layer_norm(data)
    # loss = out.sum()
    # loss.backward()



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import unittest

import torch

from protenix.model.modules.frames import (
    expressCoordinatesInFrame,
    gather_frame_atom_by_indices,
)


class TestFrame(unittest.TestCase):
    def setUp(self):
        self._start_time = time.time()

    def test_express_coordinates_in_frame(self):
        N_atom = 10
        N_frame = 5
        bs_dims = (2, 3)
        coordinates = torch.rand(size=(*bs_dims, N_atom, 3))
        frames = torch.rand(size=(*bs_dims, N_frame, 3, 3))
        x_transformed = expressCoordinatesInFrame(coordinate=coordinates, frames=frames)
        # shape check
        self.assertEqual(
            x_transformed.shape, torch.Size((*bs_dims, N_frame, N_atom, 3))
        )

        # invarient to batch order
        x_transformed12 = expressCoordinatesInFrame(
            coordinate=coordinates[1][2], frames=frames[1][2]
        )
        self.assertTrue(torch.allclose(x_transformed12, x_transformed[1][2]))

        # value check
        frames = torch.tensor([[[1, 0, 0], [0, 0, 0], [0, 1, 0]]]).float()
        coordinates = torch.tensor([[1, 0, 0]]).float()
        x_transformed = expressCoordinatesInFrame(coordinate=coordinates, frames=frames)
        self.assertTrue(
            torch.allclose(
                x_transformed,
                torch.tensor([[0.7071, -0.7071, 0.0000]]),
                atol=1e-3,
                rtol=1e-3,
            )
        )  # math.sqrt(2)/2

    def test_gather_frame_atom_by_indices(self):
        N_atom = 10
        N_frame = 5
        bs_dims = (2, 3)
        coordinates = torch.rand(size=(*bs_dims, N_atom, 3))
        indexes = torch.randint(size=(*bs_dims, N_frame, 3), low=0, high=10)
        out = gather_frame_atom_by_indices(
            coordinate=coordinates, frame_atom_index=indexes
        )
        # shape check
        self.assertEqual(out.shape, torch.Size((*bs_dims, N_frame, 3, 3)))
        coordinates = torch.rand(size=(*bs_dims, N_atom, 3))
        indexes = torch.randint(size=(N_frame, 3), low=0, high=10)
        out = gather_frame_atom_by_indices(
            coordinate=coordinates, frame_atom_index=indexes
        )
        # shape check [naive mode]
        self.assertEqual(out.shape, torch.Size((*bs_dims, N_frame, 3, 3)))

    def tearDown(self):
        elapsed_time = time.time() - self._start_time
        print(f"Test {self.id()} took {elapsed_time:.6f}s")


if __name__ == "__main__":
    unittest.main()



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Copyright 2021 AlQuraishi Laboratory

import collections
import dataclasses
import functools
import io
import re
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple

from Bio import PDB
from Bio.Data import PDBData

from protenix.data.constants import (
    DNA_CHAIN,
    PROTEIN_CHAIN,
    RNA_CHAIN,
    TEMPLATE_DNA_SEQ_TO_ID,
    TEMPLATE_PROTEIN_SEQ_TO_ID,
    TEMPLATE_RNA_SEQ_TO_ID,
)
from protenix.data.tools.common import parse_fasta
from protenix.utils.logger import get_logger

logger = get_logger(__name__)

# Type aliases for clarity.
MmCIFDict = Mapping[str, Sequence[str]]
PdbHeader = Mapping[str, Any]
ChainId = str
SeqRes = str


# --- Exceptions ---
"""Adapted from openfold.data"""


class TemplateError(Exception):
    """Base class for exceptions in this module."""


class ParsingError(TemplateError):
    """Raised when parsing mmCIF file fails or contains invalid data."""


class NoChainsError(ParsingError):
    """Raised when template mmCIF doesn't have any chains."""


class NoAtomDataInTemplateError(ParsingError):
    """Raised when template mmCIF doesn't contain atom positions."""


class TemplateAtomMaskAllZerosError(ParsingError):
    """Raised when template mmCIF had all atom positions masked."""


class MultipleChainsError(ParsingError):
    """Raised when multiple chains are found for a given ID where only one was expected."""


class AlignmentError(TemplateError):
    """Raised when alignment between query and template fails."""


class SequenceNotInTemplateError(AlignmentError):
    """Raised when template mmCIF doesn't contain the expected sequence."""


class QueryToTemplateAlignError(AlignmentError):
    """Raised when the query cannot be aligned to the template."""


class CaDistanceError(AlignmentError):
    """Raised when a CA atom distance exceeds a threshold."""


class PrefilterError(TemplateError):
    """Base class for template prefilter exceptions."""


class DateError(PrefilterError):
    """Raised when the hit date is after the max allowed date."""


class AlignRatioError(PrefilterError):
    """Raised when the hit align ratio to the query is too small."""


class DuplicateError(PrefilterError):
    """Raised when the hit is an exact subsequence of the query."""


class LengthError(PrefilterError):
    """Raised when the hit is too short."""


# --- Data Classes ---


@dataclasses.dataclass(frozen=True)
class ResiduePosition:
    """Represents a residue position in a structure."""

    chain_id: str
    residue_number: int
    insertion_code: str


@dataclasses.dataclass(frozen=True)
class ResidueAtPosition:
    """Represents a residue at a specific position, possibly missing."""

    position: Optional[ResiduePosition]
    name: str
    is_missing: bool
    hetflag: str


@dataclasses.dataclass(frozen=True)
class MmcifObject:
    """Representation of a parsed mmCIF file."""

    file_id: str
    header: PdbHeader
    structure: PDB.Structure.Structure
    chain_to_seqres: Mapping[ChainId, SeqRes]
    seqres_to_structure: Mapping[ChainId, Mapping[int, ResidueAtPosition]]
    raw_string: Any


@dataclasses.dataclass(frozen=True)
class ParsingResult:
    """Result of parsing an mmCIF file."""

    mmcif_object: Optional[MmcifObject]
    errors: Mapping[Tuple[str, str], Any]


@dataclasses.dataclass(frozen=True)
class Monomer:
    """Represents a monomer in a polymer chain."""

    id: str
    num: int


@dataclasses.dataclass(frozen=True)
class AtomSite:
    """Represents an atom site from mmCIF _atom_site loop."""

    residue_name: str
    author_chain_id: str
    mmcif_chain_id: str
    author_seq_num: str
    mmcif_seq_num: int
    insertion_code: str
    hetatm_atom: str
    model_num: str


@dataclasses.dataclass(frozen=True)
class PrefilterResult:
    """Result of pre-filtering a template hit."""

    valid: bool
    error: Optional[str] = None
    warning: Optional[str] = None


@dataclasses.dataclass(frozen=True)
class SingleHitResult:
    """Result of processing a single template hit."""

    hit: Optional["TemplateHit"]
    features: Optional[Mapping[str, Any]]
    error: Optional[str]
    warning: Optional[str]


@dataclasses.dataclass(frozen=True)
class TemplateSearchResult:
    """Result of a template search across multiple hits."""

    features: Sequence[Mapping[str, Any]]
    hits: Sequence["TemplateHit"]
    errors: Sequence[str]
    warnings: Sequence[str]


# --- mmCIF Parsing Helpers ---


def encode_template_restype(chain_type: str, sequence: str) -> List[int]:
    """Encodes a sequence of residues into integer IDs based on chain type."""
    if chain_type == PROTEIN_CHAIN:
        return [
            TEMPLATE_PROTEIN_SEQ_TO_ID.get(r, TEMPLATE_PROTEIN_SEQ_TO_ID["X"])
            for r in sequence
        ]
    if chain_type == RNA_CHAIN:
        return [
            TEMPLATE_RNA_SEQ_TO_ID.get(r, TEMPLATE_RNA_SEQ_TO_ID["N"]) for r in sequence
        ]
    if chain_type == DNA_CHAIN:
        return [
            TEMPLATE_DNA_SEQ_TO_ID.get(r, TEMPLATE_DNA_SEQ_TO_ID["N"]) for r in sequence
        ]
    raise NotImplementedError(f"Unsupported chain type: {chain_type}")


def get_pdb_id_and_chain(hit: "TemplateHit") -> Tuple[str, str]:
    """Extracts PDB ID and chain from hit name."""
    match = re.match(r"([a-zA-Z\d]{4})_([a-zA-Z0-9.]+)", hit.name)
    if not match:
        raise ValueError(f"Invalid hit name format: {hit.name}")
    return match.group(1).lower(), match.group(2)


class TemplateParser:
    """Class to parse mmCIF files into MmcifObject."""

    @staticmethod
    def _mmcif_loop_to_list(
        prefix: str, parsed_info: MmCIFDict
    ) -> List[Dict[str, str]]:
        """Extracts an mmCIF loop into a list of dictionaries."""
        cols = []
        data = []
        for key, value in parsed_info.items():
            if key.startswith(prefix):
                cols.append(key)
                data.append(value)

        if not data:
            return []

        assert all(
            len(xs) == len(data[0]) for xs in data
        ), f"mmCIF error: Not all loops are the same length for prefix {prefix}"

        return [dict(zip(cols, xs)) for xs in zip(*data)]

    @staticmethod
    def _mmcif_loop_to_dict(
        prefix: str, index_key: str, parsed_info: MmCIFDict
    ) -> Dict[str, Dict[str, str]]:
        """Extracts an mmCIF loop into a dictionary keyed by a specific column."""
        entries = TemplateParser._mmcif_loop_to_list(prefix, parsed_info)
        return {entry[index_key]: entry for entry in entries}

    @staticmethod
    def _get_release_date(parsed_info: MmCIFDict) -> str:
        """Returns the oldest revision date from mmCIF data."""
        revision_dates = parsed_info.get("_pdbx_audit_revision_history.revision_date")
        if revision_dates:
            return min(revision_dates)
        return ""

    @staticmethod
    def _get_header(parsed_info: MmCIFDict) -> PdbHeader:
        """Extracts basic header information (method, release date, resolution)."""
        header = {}
        experiments = TemplateParser._mmcif_loop_to_list("_exptl.", parsed_info)
        header["structure_method"] = ",".join(
            [exp.get("_exptl.method", "").lower() for exp in experiments]
        )

        release_date = TemplateParser._get_release_date(parsed_info)
        if release_date:
            header["release_date"] = release_date
        else:
            logger.warning(
                "Could not determine release_date for %s", parsed_info.get("_entry.id")
            )

        header["resolution"] = 0.00
        for res_key in (
            "_refine.ls_d_res_high",
            "_em_3d_reconstruction.resolution",
            "_reflns.d_resolution_high",
        ):
            if res_key in parsed_info:
                try:
                    header["resolution"] = float(parsed_info[res_key][0])
                    break
                except (ValueError, IndexError):
                    continue
        return header

    @staticmethod
    def _get_first_model(structure: PDB.Structure.Structure) -> PDB.Model.Model:
        """Returns the first model in a Biopython structure."""
        return next(structure.get_models())

    @staticmethod
    def _get_atom_site_list(parsed_info: MmCIFDict) -> List[AtomSite]:
        """Returns a list of AtomSite objects from mmCIF data."""
        return [
            AtomSite(*site)
            for site in zip(
                parsed_info["_atom_site.label_comp_id"],
                parsed_info["_atom_site.auth_asym_id"],
                parsed_info["_atom_site.label_asym_id"],
                parsed_info["_atom_site.auth_seq_id"],
                parsed_info["_atom_site.label_seq_id"],
                parsed_info["_atom_site.pdbx_PDB_ins_code"],
                parsed_info["_atom_site.group_PDB"],
                parsed_info["_atom_site.pdbx_PDB_model_num"],
            )
        ]

    @staticmethod
    def _get_protein_chains(parsed_info: MmCIFDict) -> Dict[ChainId, Sequence[Monomer]]:
        """Extracts valid protein chains and their sequences."""
        entity_poly_seqs = TemplateParser._mmcif_loop_to_list(
            "_entity_poly_seq.", parsed_info
        )
        polymers = collections.defaultdict(list)
        for entry in entity_poly_seqs:
            polymers[entry["_entity_poly_seq.entity_id"]].append(
                Monomer(
                    id=entry["_entity_poly_seq.mon_id"],
                    num=int(entry["_entity_poly_seq.num"]),
                )
            )

        chem_comps = TemplateParser._mmcif_loop_to_dict(
            "_chem_comp.", "_chem_comp.id", parsed_info
        )
        struct_asyms = TemplateParser._mmcif_loop_to_list("_struct_asym.", parsed_info)

        entity_to_mmcif_chains = collections.defaultdict(list)
        for asym in struct_asyms:
            entity_to_mmcif_chains[asym["_struct_asym.entity_id"]].append(
                asym["_struct_asym.id"]
            )

        valid_chains = {}
        for entity_id, seq_info in polymers.items():
            # Check if any component is peptide-like.
            is_protein = any(
                "peptide" in chem_comps.get(m.id, {}).get("_chem_comp.type", "")
                for m in seq_info
            )
            if is_protein:
                for chain_id in entity_to_mmcif_chains[entity_id]:
                    valid_chains[chain_id] = seq_info
        return valid_chains

    @staticmethod
    def _is_set(data: str) -> bool:
        """Checks if mmCIF data is set (not '.' or '?')."""
        return data not in (".", "?")

    @staticmethod
    @functools.lru_cache(maxsize=16)
    def parse(
        *,
        file_id: str,
        mmcif_string: str,
        auth_chain_id: Optional[str] = None,
        catch_all_errors: bool = True,
    ) -> ParsingResult:
        """Parses an mmCIF string into an MmcifObject."""
        errors = {}
        try:
            parser = PDB.MMCIFParser(QUIET=True)
            structure = parser.get_structure("", io.StringIO(mmcif_string))
            first_model = TemplateParser._get_first_model(structure)
            parsed_info = parser._mmcif_dict  # pylint: disable=protected-access

            # Ensure all values are lists.
            for k, v in parsed_info.items():
                if not isinstance(v, list):
                    parsed_info[k] = [v]

            header = TemplateParser._get_header(parsed_info)
            valid_chains = TemplateParser._get_protein_chains(parsed_info)
            if not valid_chains:
                return ParsingResult(None, {(file_id, ""): "No protein chains found."})

            seq_start_num = {
                cid: min(m.num for m in seq) for cid, seq in valid_chains.items()
            }
            mmcif_to_author_chain_id = {}
            seq_to_structure_mappings = collections.defaultdict(dict)

            for atom in TemplateParser._get_atom_site_list(parsed_info):
                if atom.model_num != "1":
                    continue
                if auth_chain_id is not None and atom.author_chain_id != auth_chain_id:
                    continue

                mmcif_to_author_chain_id[atom.mmcif_chain_id] = atom.author_chain_id

                if atom.mmcif_chain_id in valid_chains:
                    hetflag = " "
                    if atom.hetatm_atom == "HETATM":
                        hetflag = (
                            "W"
                            if atom.residue_name in ("HOH", "WAT")
                            else f"H_{atom.residue_name}"
                        )

                    ins_code = (
                        atom.insertion_code
                        if TemplateParser._is_set(atom.insertion_code)
                        else " "
                    )
                    pos = ResiduePosition(
                        chain_id=atom.author_chain_id,
                        residue_number=int(atom.author_seq_num),
                        insertion_code=ins_code,
                    )
                    seq_idx = (
                        int(atom.mmcif_seq_num) - seq_start_num[atom.mmcif_chain_id]
                    )
                    seq_to_structure_mappings[atom.author_chain_id][
                        seq_idx
                    ] = ResidueAtPosition(
                        position=pos,
                        name=atom.residue_name,
                        is_missing=False,
                        hetflag=hetflag,
                    )

            # Fill in missing residues.
            for cid, seq_info in valid_chains.items():
                if cid not in mmcif_to_author_chain_id:
                    continue
                auth_cid = mmcif_to_author_chain_id[cid]
                mapping = seq_to_structure_mappings[auth_cid]
                for idx, monomer in enumerate(seq_info):
                    if idx not in mapping:
                        mapping[idx] = ResidueAtPosition(
                            position=None, name=monomer.id, is_missing=True, hetflag=" "
                        )

            # Convert 3-letter to 1-letter sequences.
            auth_chain_to_seq = {}
            for cid, seq_info in valid_chains.items():
                if cid not in mmcif_to_author_chain_id:
                    continue
                auth_cid = mmcif_to_author_chain_id[cid]
                seq = "".join(
                    [
                        PDBData.protein_letters_3to1.get(monomer.id, "X")
                        for monomer in seq_info
                    ]
                )
                # Ensure it's 1-letter.
                seq = "".join([c if len(c) == 1 else "X" for c in seq])
                auth_chain_to_seq[auth_cid] = seq

            mmcif_obj = MmcifObject(
                file_id=file_id,
                header=header,
                structure=first_model,
                chain_to_seqres=auth_chain_to_seq,
                seqres_to_structure=seq_to_structure_mappings,
                raw_string=parsed_info,
            )
            return ParsingResult(mmcif_object=mmcif_obj, errors=errors)
        except Exception as e:
            errors[(file_id, "")] = e
            if not catch_all_errors:
                raise
            return ParsingResult(None, errors=errors)


@dataclasses.dataclass(frozen=True)
class TemplateHit:
    """Represents a template hit from a search tool (e.g., HHSearch, hmmsearch)."""

    index: int
    name: str
    aligned_cols: int
    sum_probs: Optional[float]
    query: str
    hit_sequence: str
    indices_query: List[int]
    indices_hit: List[int]

    @functools.cached_property
    def query_to_hit_mapping(self) -> Mapping[int, int]:
        """Maps 0-based query indices to 0-based hit indices."""
        mapping = {}
        for q_idx, h_idx in zip(self.indices_query, self.indices_hit):
            if (q_idx != -1) and (h_idx != -1):
                mapping[q_idx] = h_idx
        return mapping


@dataclasses.dataclass(frozen=True)
class HitMetadata:
    """Metadata parsed from an hmmsearch A3M description line."""

    pdb_id: str
    chain: str
    start: int
    end: int
    length: int
    text: str


class HHRParser:
    """Class to parse HHR files from HHSearch."""

    @staticmethod
    def parse(hhr_string: str) -> List[TemplateHit]:
        """
        Parses an entire HHR file content.

        Args:
            hhr_string: The content of the HHR file.

        Returns:
            A list of TemplateHit objects.
        """
        lines = hhr_string.splitlines()
        block_starts = [i for i, line in enumerate(lines) if line.startswith("No ")]
        hits = []
        if block_starts:
            block_starts.append(len(lines))
            for i in range(len(block_starts) - 1):
                hits.append(
                    HHRParser._parse_hit(lines[block_starts[i] : block_starts[i + 1]])
                )
        return hits

    @staticmethod
    def _parse_hit(lines: Sequence[str]) -> TemplateHit:
        """Parses a single hit block from an HHR file."""
        hit_num = int(lines[0].split()[-1])
        hit_name = lines[1][1:].strip()
        summary = lines[2]
        match = re.search(r"Aligned_cols=(\d+).*Sum_probs=([0-9.]+)", summary)
        if not match:
            raise RuntimeError(f"Could not parse HHR summary: {summary}")
        cols, probs = int(match.group(1)), float(match.group(2))

        query, hit_seq = "", ""
        idx_q, idx_h = [], []
        for line in lines[3:]:
            if line.startswith("Q ") and not any(
                line.startswith(x) for x in ("Q ss_", "Q Consensus")
            ):
                match = re.search(r"\s+(\d+)\s+([A-Z-]+)\s+(\d+)", line[17:])
                if match:
                    start = int(match.group(1)) - 1
                    seq = match.group(2)
                    query += seq
                    HHRParser._update_residue_indices(seq, start, idx_q)
            elif line.startswith("T ") and not any(
                line.startswith(x) for x in ("T ss_", "T Consensus")
            ):
                match = re.search(r"\s+(\d+)\s+([A-Z-]+)", line[17:])
                if match:
                    start = int(match.group(1)) - 1
                    seq = match.group(2)
                    hit_seq += seq
                    HHRParser._update_residue_indices(seq, start, idx_h)

        return TemplateHit(hit_num, hit_name, cols, probs, query, hit_seq, idx_q, idx_h)

    @staticmethod
    def _update_residue_indices(seq: str, start: int, indices: List[int]):
        """Updates the list of residue indices for a sequence segment."""
        curr = start
        for char in seq:
            if char == "-":
                indices.append(-1)
            else:
                indices.append(curr)
                curr += 1


class HmmsearchA3MParser:
    """Class to parse A3M files from hmmsearch."""

    @staticmethod
    def parse(
        query_seq: str, a3m_str: str, skip_first: bool = True
    ) -> List[TemplateHit]:
        """
        Parses an A3M string from hmmsearch.

        Args:
            query_seq: The query sequence.
            a3m_str: The content of the A3M file.
            skip_first: Whether to skip the first sequence (usually the query).

        Returns:
            A list of TemplateHit objects.
        """
        seqs, descs = parse_fasta(a3m_str)
        parsed = list(zip(seqs, descs))
        if skip_first:
            parsed = parsed[1:]

        idx_q = HmmsearchA3MParser._get_indices(query_seq, 0)
        hits = []
        for i, (h_seq, h_desc) in enumerate(parsed, start=1):
            if "mol:protein" not in h_desc:
                continue
            meta = HmmsearchA3MParser._parse_description(h_desc)
            cols = sum(1 for r in h_seq if r.isupper() and r != "-")
            idx_h = HmmsearchA3MParser._get_indices(h_seq, meta.start - 1)
            hits.append(
                TemplateHit(
                    i,
                    f"{meta.pdb_id}_{meta.chain}",
                    cols,
                    None,
                    query_seq,
                    h_seq.upper(),
                    idx_q,
                    idx_h,
                )
            )
        return hits

    @staticmethod
    def _get_indices(seq: str, start: int) -> List[int]:
        """Calculates residue indices for a sequence with gaps or insertions."""
        indices = []
        curr = start
        for char in seq:
            if char == "-":
                indices.append(-1)
            elif char.islower():
                curr += 1
            else:
                indices.append(curr)
                curr += 1
        return indices

    @staticmethod
    def _parse_description(desc: str) -> HitMetadata:
        """Parses the description line from an hmmsearch hit."""
        pattern = (
            r"^>?([a-z0-9]+)_(\w+)/([0-9]+)-([0-9]+).*protein length:([0-9]+) *(.*)$"
        )
        match = re.match(pattern, desc.strip())
        if not match:
            raise ValueError(f"Could not parse hmmsearch description: {desc}")
        return HitMetadata(
            match[1], match[2], int(match[3]), int(match[4]), int(match[5]), match[6]
        )



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2025 Shad Nygren, Virtual Hipster Corporation
# Contributed to the Protenix project under the Apache License 2.0
"""
Test suite for ESM model loading compatibility.
Related to GitHub issue #176 (ESM weights loading error with PyTorch 2.6+)
"""

import unittest


class TestESMLoading(unittest.TestCase):
    """Test ESM model loading with different PyTorch versions"""

    def test_pytorch_version(self):
        """Check PyTorch version and document compatibility"""
        try:
            import torch

            version = torch.__version__
            major, minor = version.split(".")[:2]
            major, minor = int(major), int(
                minor.split("+")[0]
            )  # Handle versions like "2.6.0+cpu"

            if major == 2 and minor >= 6:
                print(
                    f"PyTorch {version} detected - may have ESM loading issues (issue #176)"
                )
            else:
                print(f"PyTorch {version} should work with ESM")

        except ImportError:
            self.skipTest("PyTorch not installed")

    def test_esm_import(self):
        """Test basic ESM package import"""
        try:
            import esm  # noqa: F401

            self.assertTrue(True, "ESM package imported successfully")
        except ImportError:
            self.skipTest("ESM (fair-esm) not installed")

    def test_esm_model_loading(self):
        """
        Test for issue #176: ESM weights loading error in PyTorch 2.6+
        The error occurs when loading pretrained ESM models
        """
        try:
            import torch

            version = torch.__version__
            major, minor = version.split(".")[:2]
            major, minor = int(major), int(minor.split("+")[0])

            # Try to load ESM model (this is where issue #176 occurs)
            from protenix.data.esm.compute_esm import get_esm_embedding  # noqa: F401

            # If we're on PyTorch 2.6+ and this works, the bug is fixed
            if major == 2 and minor >= 6:
                print("ESM loading works on PyTorch 2.6+ - issue #176 may be resolved")

            self.assertTrue(True)

        except ImportError as e:
            if "compute_esm" in str(e):
                self.skipTest("Protenix not installed")
            elif "torch" in str(e):
                self.skipTest("PyTorch not installed")
            else:
                # This might be the ESM loading issue
                try:
                    import torch

                    version = torch.__version__
                    if version.startswith("2.6") or version.startswith("2.7"):
                        self.fail(
                            f"ESM loading failed on PyTorch {version}. "
                            "This is issue #176. The fix has been merged - update Protenix."
                        )
                    else:
                        self.skipTest(f"ESM loading failed: {e}")
                except ImportError:
                    self.skipTest("PyTorch not installed")
        except Exception as e:
            self.skipTest(f"Cannot test ESM loading: {e}")

    def test_esm_embedding_generation(self):
        """Test that ESM embeddings can be generated for a sequence"""
        try:
            from protenix.data.esm.compute_esm import get_esm_embedding

            # Test with a small protein sequence
            test_sequence = "MKTAYDELAAEAFLEENTPILH"

            # This should work if ESM is properly loaded
            embedding = get_esm_embedding(test_sequence)

            # Check that we got a reasonable embedding
            self.assertIsNotNone(embedding)
            self.assertGreater(len(embedding), 0)

        except ImportError:
            self.skipTest("Protenix or ESM not installed")
        except Exception as e:
            # Check if this is the PyTorch 2.6+ issue
            import torch

            if torch.__version__.startswith("2.6") or torch.__version__.startswith(
                "2.7"
            ):
                self.fail(
                    f"ESM embedding generation failed on PyTorch {torch.__version__}: {e}"
                )
            else:
                self.skipTest(f"Cannot test ESM embeddings: {e}")


if __name__ == "__main__":
    unittest.main()



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from typing import Any

from torch.utils.cpp_extension import load


def compile(
    name: str, sources: list[str], extra_include_paths: list[str], build_directory: str
) -> Any:
    os.environ["TORCH_CUDA_ARCH_LIST"] = "7.0;8.0"
    return load(
        name=name,
        sources=sources,
        extra_include_paths=extra_include_paths,
        extra_cflags=[
            "-O3",
            "-DVERSION_GE_1_1",
            "-DVERSION_GE_1_3",
            "-DVERSION_GE_1_5",
        ],
        extra_cuda_cflags=[
            "-O3",
            "--use_fast_math",
            "-DVERSION_GE_1_1",
            "-DVERSION_GE_1_3",
            "-DVERSION_GE_1_5",
            "-std=c++17",
            "-maxrregcount=32",
            "-U__CUDA_NO_HALF_OPERATORS__",
            "-U__CUDA_NO_HALF_CONVERSIONS__",
            "--expt-relaxed-constexpr",
            "--expt-extended-lambda",
            "-gencode",
            "arch=compute_70,code=sm_70",
            "-gencode",
            "arch=compute_80,code=sm_80",
            "-gencode",
            "arch=compute_86,code=sm_86",
            "-gencode",
            "arch=compute_90,code=sm_90",
        ],
        verbose=True,
        build_directory=build_directory,
    )



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2025 Shad Nygren, Virtual Hipster Corporation
# Contributed to the Protenix project under the Apache License 2.0

"""
Test suite for Triton kernel compatibility.
Related to GitHub issue #185 (Triton version compatibility)
"""

import os
import unittest


class TestTritonCompatibility(unittest.TestCase):
    """Test Triton kernel compatibility across different GPU types"""

    def test_triton_version(self):
        """Verify Triton version is compatible (3.3.0 recommended)"""
        try:
            import triton

            version = triton.__version__
            # Issue #185 recommends triton==3.3.0
            self.assertEqual(
                version.split(".")[0],
                "3",
                f"Triton major version should be 3, got {version}",
            )
        except ImportError:
            self.skipTest("Triton not installed")

    def test_triton_kernel_import(self):
        """
        Test for issue #185: Triton kernel import failure
        Error: "Failed to import Triton-based component: triangle_multiplicative_update: Not Supported"
        """
        try:
            # This import fails on RTX 3090/4090 GPUs
            from protenix.model.tri_attention import op  # noqa: F401

            self.assertTrue(True, "Triton attention kernel imported successfully")
        except ImportError as e:
            if "Not Supported" in str(e):
                # This is the bug from issue #185
                import torch

                if torch.cuda.is_available():
                    gpu_name = torch.cuda.get_device_name()
                    if "3090" in gpu_name or "4090" in gpu_name:
                        self.fail(
                            f"Triton not supported on {gpu_name}. "
                            "This is issue #185. Use --trimul_kernel 'torch' as workaround."
                        )
                # For other GPUs, this might be expected
                self.skipTest(f"Triton kernel not supported: {e}")
            else:
                # Different import error
                self.skipTest(f"Protenix not installed: {e}")
        except Exception as e:
            # Unexpected error
            self.skipTest(f"Cannot test Triton compatibility: {e}")

    def test_torch_fallback(self):
        """Verify torch fallback works when Triton is not available"""
        # Set environment to use torch kernel
        os.environ["TRIMUL_KERNEL"] = "torch"

        try:
            # This should work even without Triton
            import protenix.model.modules  # noqa: F401

            self.assertTrue(True, "Torch fallback kernel works")
        except ImportError:
            self.skipTest("Protenix not installed")
        finally:
            # Clean up environment
            if "TRIMUL_KERNEL" in os.environ:
                del os.environ["TRIMUL_KERNEL"]

    def test_gpu_compatibility_check(self):
        """Document which GPUs have Triton compatibility issues"""
        try:
            import torch

            if not torch.cuda.is_available():
                self.skipTest("No GPU available")

            gpu_name = torch.cuda.get_device_name()
            known_issues = [
                "RTX 3090",
                "RTX 4090",
                "GeForce RTX 3090",
                "GeForce RTX 4090",
            ]

            for gpu in known_issues:
                if gpu in gpu_name:
                    print(
                        f"Warning: {gpu_name} has known Triton compatibility issues (issue #185)"
                    )
                    print("Workaround: Use --trimul_kernel 'torch'")
                    # Don't fail the test, just document it
                    return

            print(f"GPU {gpu_name} should work with Triton kernels")

        except ImportError:
            self.skipTest("PyTorch not installed")


if __name__ == "__main__":
    unittest.main()



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import biotite.structure as struc
import numpy as np
from biotite.structure import AtomArray, get_molecule_indices
from scipy.spatial.distance import cdist

from protenix.data.constants import CRYSTALLIZATION_AIDS


class Filter(object):
    """
    Ref: AlphaFold3 SI Chapter 2.5.4
    """

    @staticmethod
    def remove_hydrogens(atom_array: AtomArray) -> AtomArray:
        """remove hydrogens and deuteriums"""
        return atom_array[~np.isin(atom_array.element, ["H", "D"])]

    @staticmethod
    def remove_water(atom_array: AtomArray) -> AtomArray:
        """remove water (HOH) and deuterated water (DOD)"""
        return atom_array[~np.isin(atom_array.res_name, ["HOH", "DOD"])]

    @staticmethod
    def remove_element_X(atom_array: AtomArray) -> AtomArray:
        """
        remove element X
        following residues have element X:
        - UNX: unknown one atom or ion
        - UNL: unknown ligand, some atoms are marked as X
        - ASX: ASP/ASN ambiguous, two ambiguous atoms are marked as X, 6 entries in the PDB
        - GLX: GLU/GLN ambiguous, two ambiguous atoms are marked as X, 5 entries in the PDB
        """
        X_mask = np.zeros(len(atom_array), dtype=bool)
        starts = struc.get_residue_starts(atom_array, add_exclusive_stop=True)
        for start, stop in zip(starts[:-1], starts[1:]):
            res_name = atom_array.res_name[start]
            if res_name in ["UNX", "UNL"]:
                X_mask[start:stop] = True
        atom_array = atom_array[~X_mask]

        # map ASX to ASP, as ASP is more symmetric than ASN
        mask = atom_array.res_name == "ASX"
        atom_array.res_name[mask] = "ASP"
        atom_array.atom_name[mask & (atom_array.atom_name == "XD1")] = "OD1"
        atom_array.atom_name[mask & (atom_array.atom_name == "XD2")] = "OD2"
        atom_array.element[mask & (atom_array.element == "X")] = "O"

        # map GLX to GLU, as GLU is more symmetric than GLN
        mask = atom_array.res_name == "GLX"
        atom_array.res_name[mask] = "GLU"
        atom_array.atom_name[mask & (atom_array.atom_name == "XE1")] = "OE1"
        atom_array.atom_name[mask & (atom_array.atom_name == "XE2")] = "OE2"
        atom_array.element[mask & (atom_array.element == "X")] = "O"
        return atom_array

    @staticmethod
    def remove_crystallization_aids(
        atom_array: AtomArray, entity_poly_type: dict
    ) -> AtomArray:
        """remove crystallization aids, eg: SO4, GOL, etc.

        Only remove crystallization aids if the chain is not polymer.

        Ref: AlphaFold3 SI Chapter 2.5.4
        """
        non_aids_mask = ~np.isin(atom_array.res_name, CRYSTALLIZATION_AIDS)
        poly_mask = np.isin(atom_array.label_entity_id, list(entity_poly_type.keys()))
        return atom_array[poly_mask | non_aids_mask]

    @staticmethod
    def _get_clashing_chains(
        atom_array: AtomArray, chain_ids: list[str]
    ) -> tuple[np.ndarray, list[int]]:
        """
        Calculate the number of atoms clashing with other chains for each chain
        and return a matrix that records the count of clashing atoms.

        Note: if two chains are covalent, they are not considered as clashing.

        Args:
            atom_array (AtomArray): All atoms, including those not resolved.
            chain_ids (list[str]): Unique chain indices of resolved atoms.

        Returns:
            tuple:
                clash_records (numpy.ndarray): Matrix of clashing atom num.
                                               (i, j) means the ratio of i's atom clashed with j's atoms.
                                               Note: (i, j) != (j, i).
                chain_resolved_atom_nums (list[int]): The number of resolved atoms corresponding to each chain ID.
        """
        is_resolved_centre_atom = (
            atom_array.centre_atom_mask == 1
        ) & atom_array.is_resolved
        cell_list = struc.CellList(
            atom_array, cell_size=1.7, selection=is_resolved_centre_atom
        )

        # (i, j) means the ratio of i's atom clashed with j's atoms
        clash_records = np.zeros((len(chain_ids), len(chain_ids)))

        # record the number of resolved atoms for each chain
        chain_resolved_atom_nums = []

        # record covalent relationship between chains
        chains_covalent_dict = {}
        for idx, chain_id_i in enumerate(chain_ids):
            for chain_id_j in chain_ids[idx + 1 :]:
                mol_indices = get_molecule_indices(
                    atom_array[np.isin(atom_array.chain_id, [chain_id_i, chain_id_j])]
                )
                if len(mol_indices) == 1:
                    covalent = 1
                else:
                    covalent = 0
                chains_covalent_dict[(chain_id_i, chain_id_j)] = covalent
                chains_covalent_dict[(chain_id_j, chain_id_i)] = covalent

        for i, chain_id in enumerate(chain_ids):
            coords = atom_array.coord[
                (atom_array.chain_id == chain_id) & is_resolved_centre_atom
            ]
            chain_resolved_atom_nums.append(len(coords))
            chain_atom_ids = np.where(atom_array.chain_id == chain_id)[0]
            chain_atom_ids_set = set(chain_atom_ids) | {-1}

            # Get atom indices from the current cell and the eight surrounding cells.
            neighbors_ids_2d = cell_list.get_atoms_in_cells(coords, cell_radius=1)
            neighbors_ids = np.unique(neighbors_ids_2d)

            # Remove the atom indices of the current chain.
            other_chain_atom_ids = list(set(neighbors_ids) - chain_atom_ids_set)

            if not other_chain_atom_ids:
                continue
            else:
                # Calculate the distance matrix with neighboring atoms.
                other_chain_atom_coords = atom_array.coord[other_chain_atom_ids]
                dist_mat = cdist(coords, other_chain_atom_coords, metric="euclidean")
                clash_mat = dist_mat < 1.6  # change 1.7 to 1.6 for more compatibility
                if np.any(clash_mat):
                    clashed_other_chain_ids = atom_array.chain_id[other_chain_atom_ids]

                    for other_chain_id in set(clashed_other_chain_ids):

                        # two chains covalent with each other
                        if chains_covalent_dict[(chain_id, other_chain_id)]:
                            continue

                        cols = np.where(clashed_other_chain_ids == other_chain_id)[0]

                        # how many i's atoms clashed with j
                        any_atom_clashed = np.any(
                            clash_mat[:, cols].astype(int), axis=1
                        )
                        clashed_atom_num = np.sum(any_atom_clashed.astype(int))

                        if clashed_atom_num > 0:
                            j = chain_ids.index(other_chain_id)
                            clash_records[i][j] += clashed_atom_num
        return clash_records, chain_resolved_atom_nums

    @staticmethod
    def _get_removed_clash_chain_ids(
        clash_records: np.ndarray,
        chain_ids: list[str],
        chain_resolved_atom_nums: list[int],
        core_chain_id: np.ndarray = [],
    ) -> list[str]:
        """
        Perform pairwise comparisons on the chains, and select the chain IDs
        to be deleted according to the clahsing chain rules.

        Args:
            clash_records (numpy.ndarray): Matrix of clashing atom num.
                                           (i, j) means the ratio of i's atom clashed with j's atoms.
                                           Note: (i, j) != (j, i).
            chain_ids (list[str]): Unique chain indices of resolved atoms.
            chain_resolved_atom_nums (list[int]): The number of resolved atoms corresponding to each chain ID.
            core_chain_id (np.ndarray): The chain ID of the core chain.

        Returns:
            list[str]: A list of chain IDs that have been determined for deletion.
        """
        removed_chain_ids = []
        for i in range(len(chain_ids)):
            atom_num_i = chain_resolved_atom_nums[i]
            chain_idx_i = chain_ids[i]

            if chain_idx_i in removed_chain_ids:
                continue

            for j in range(i + 1, len(chain_ids)):
                atom_num_j = chain_resolved_atom_nums[j]
                chain_idx_j = chain_ids[j]

                if chain_idx_j in removed_chain_ids:
                    continue

                clash_num_ij, clash_num_ji = (
                    clash_records[i][j],
                    clash_records[j][i],
                )

                clash_ratio_ij = clash_num_ij / atom_num_i
                clash_ratio_ji = clash_num_ji / atom_num_j

                if clash_ratio_ij <= 0.3 and clash_ratio_ji <= 0.3:
                    # not reaches the threshold
                    continue
                else:
                    # clashing chains
                    if (
                        chain_idx_i in core_chain_id
                        and chain_idx_j not in core_chain_id
                    ):
                        removed_chain_idx = chain_idx_j
                    elif (
                        chain_idx_i not in core_chain_id
                        and chain_idx_j in core_chain_id
                    ):
                        removed_chain_idx = chain_idx_i

                    elif clash_ratio_ij > clash_ratio_ji:
                        removed_chain_idx = chain_idx_i
                    elif clash_ratio_ij < clash_ratio_ji:
                        removed_chain_idx = chain_idx_j
                    else:
                        if atom_num_i < atom_num_j:
                            removed_chain_idx = chain_idx_i
                        elif atom_num_i > atom_num_j:
                            removed_chain_idx = chain_idx_j
                        else:
                            removed_chain_idx = sorted([chain_idx_i, chain_idx_j])[1]

                    removed_chain_ids.append(removed_chain_idx)

                    if removed_chain_idx == chain_idx_i:
                        # chain i already removed
                        break
        return removed_chain_ids

    @staticmethod
    def remove_polymer_chains_all_residues_unknown(
        atom_array: AtomArray,
        entity_poly_type: dict,
    ) -> AtomArray:
        """remove chains with all residues unknown"""
        chain_starts = struc.get_chain_starts(atom_array, add_exclusive_stop=True)
        invalid_chains = []  # list of [start, end)
        for index in range(len(chain_starts) - 1):
            start, end = chain_starts[index], chain_starts[index + 1]
            entity_id = atom_array[start].label_entity_id
            if (
                entity_poly_type.get(entity_id, "non-poly") == "polypeptide(L)"
                and np.all(atom_array.res_name[start:end] == "UNK")
            ) or (
                entity_poly_type.get(entity_id, "non-poly")
                in (
                    "polyribonucleotide",
                    "polydeoxyribonucleotide",
                )
                and np.all(atom_array.res_name[start:end] == "N")
            ):
                invalid_chains.append((start, end))
        mask = np.ones(len(atom_array), dtype=bool)
        for start, end in invalid_chains:
            mask[start:end] = False
        atom_array = atom_array[mask]
        return atom_array

    @staticmethod
    def remove_polymer_chains_too_short(
        atom_array: AtomArray, entity_poly_type: dict
    ) -> AtomArray:
        chain_starts = struc.get_chain_starts(atom_array, add_exclusive_stop=True)
        invalid_chains = []  # list of [start, end)
        for index in range(len(chain_starts) - 1):
            start, end = chain_starts[index], chain_starts[index + 1]
            entity_id = atom_array[start].label_entity_id
            num_residue_ids = len(set(atom_array.label_seq_id[start:end]))
            if (
                entity_poly_type.get(entity_id, "non-poly")
                in (
                    "polypeptide(L)",
                    "polyribonucleotide",
                    "polydeoxyribonucleotide",
                )
                and num_residue_ids < 4
            ):
                invalid_chains.append((start, end))
        mask = np.ones(len(atom_array), dtype=bool)
        for start, end in invalid_chains:
            mask[start:end] = False
        atom_array = atom_array[mask]
        return atom_array

    @staticmethod
    def remove_polymer_chains_with_consecutive_c_alpha_too_far_away(
        atom_array: AtomArray, entity_poly_type: dict, max_distance: float = 10.0
    ) -> AtomArray:
        chain_starts = struc.get_chain_starts(atom_array, add_exclusive_stop=True)
        invalid_chains = []  # list of [start, end)
        for index in range(len(chain_starts) - 1):
            start, end = chain_starts[index], chain_starts[index + 1]
            entity_id = atom_array.label_entity_id[start]
            if entity_poly_type.get(entity_id, "non-poly") == "polypeptide(L)":
                peptide_atoms = atom_array[start:end]
                ca_atoms = peptide_atoms[peptide_atoms.atom_name == "CA"]
                seq_ids = ca_atoms.label_seq_id
                seq_ids[seq_ids == "."] = "-100"
                seq_ids = seq_ids.astype(np.int64)
                dist_square = np.sum(
                    (ca_atoms[:-1].coord - ca_atoms[1:].coord) ** 2, axis=-1
                )
                invalid_neighbor_mask = (dist_square > max_distance**2) & (
                    seq_ids[:-1] + 1 == seq_ids[1:]
                )
                if np.any(invalid_neighbor_mask):
                    invalid_chains.append((start, end))
        mask = np.ones(len(atom_array), dtype=bool)
        for start, end in invalid_chains:
            mask[start:end] = False
        atom_array = atom_array[mask]
        return atom_array

    @staticmethod
    def too_many_chains_filter(
        atom_array: AtomArray,
        interface_radius: int = 15,
        max_chains_num: int = 20,
        core_indices: list[int] = None,
        max_tokens_num: int = None,
    ) -> tuple[AtomArray, int]:
        """
        Ref: AlphaFold3 SI Chapter 2.5.4

        For bioassemblies with greater than 20 chains, we select a random interface token
        (with a centre atom <15 Å to the centre atom of a token in another chain)
        and select the closest 20 chains to this token based on
        minimum distance between any tokens centre atom.

        Note: due to the presence of covalent small molecules,
        treat the covalent small molecule and the polymer it is attached to
        as a single chain to avoid inadvertently removing the covalent small molecules.
        Use the mol_id added to the AtomArray to differentiate between the various
        parts of the structure composed of covalent bonds.

        Args:
            atom_array (AtomArray): Biotite AtomArray Object of a Bioassembly.
            interface_radius (int, optional): Atoms within this distance of the central atom are considered interface atoms.
                                            Defaults to 15.
            max_chains_num (int, optional): The maximum number of chains permitted in a bioassembly.
                                            Filtration will be applied if exceeds this value. Defaults to 20.
            core_indices (list[int], optional): A list of indices to be used as chose the central atom.
                                                     And corresponding chains in the list will be selected proriority.
                                                     If None, a random index from whole AtomArray will be selected. Defaults to None.
            max_tokens_num (int, optional): The maximum number of tokens permitted in a bioassembly.
                                            If not None,  after more than max_chains_num, if the max_tokens_num is not reached,
                                            it will continue to append the chains.

        Returns:
            tuple:
                - atom_array (AtomArray): An AtomArray that has been processed through this filter.
                - input_chains_num (int): The number of chain in the input AtomArray.
                                          This is to log whether the filter has been utilized.
        """
        # each mol is a so called "chain" in the context of this filter.
        input_chains_num = len(np.unique(atom_array.mol_id))
        if input_chains_num <= max_chains_num:
            # no change
            return atom_array, input_chains_num

        is_resolved_centre_atom = (
            atom_array.centre_atom_mask == 1
        ) & atom_array.is_resolved

        cell_list = struc.CellList(
            atom_array, cell_size=interface_radius, selection=is_resolved_centre_atom
        )
        resolved_centre_atom = atom_array[is_resolved_centre_atom]

        assert resolved_centre_atom, "There is no resolved central atom."

        # random pick centre atom
        if core_indices is None:
            index_shuf = np.random.default_rng(seed=42).permutation(
                len(resolved_centre_atom)
            )
        else:
            index_shuf = np.array(core_indices)
            resolved_centre_atom_indices = np.nonzero(is_resolved_centre_atom)[0]

            # get indices of resolved_centre_atom
            index_shuf = np.array(
                [
                    np.where(resolved_centre_atom_indices == idx)[0][0]
                    for idx in index_shuf
                    if idx in resolved_centre_atom_indices
                ]
            )
            np.random.default_rng(seed=42).shuffle(index_shuf)

        chosen_centre_atom = None
        for idx in index_shuf:
            centre_atom = resolved_centre_atom[idx]
            neighbors_indices = cell_list.get_atoms(
                centre_atom.coord, radius=interface_radius
            )
            neighbors_indices = neighbors_indices[neighbors_indices != -1]

            neighbors_chain_ids = np.unique(atom_array.mol_id[neighbors_indices])
            # neighbors include centre atom itself
            if len(neighbors_chain_ids) > 1:
                chosen_centre_atom = centre_atom
                break

        # The distance between the central atoms in any two chains is greater than 15 angstroms.
        if chosen_centre_atom is None:
            return None, input_chains_num

        dist_mat = cdist(centre_atom.coord.reshape((1, -1)), resolved_centre_atom.coord)
        sorted_chain_id = np.array(
            [
                chain_id
                for chain_id, _dist in sorted(
                    zip(resolved_centre_atom.mol_id, dist_mat[0]),
                    key=lambda pair: pair[1],
                )
            ]
        )

        if core_indices is not None:
            # select core proriority
            core_mol_id = np.unique(atom_array.mol_id[core_indices])
            in_core_mask = np.isin(sorted_chain_id, core_mol_id)
            sorted_chain_id = np.concatenate(
                (sorted_chain_id[in_core_mask], sorted_chain_id[~in_core_mask])
            )

        closest_chain_id = set()
        chain_ids_to_token_num = {}
        if max_tokens_num is None:
            max_tokens_num = 0

        tokens = 0
        for chain_id in sorted_chain_id:
            # get token num
            if chain_id not in chain_ids_to_token_num:
                chain_ids_to_token_num[chain_id] = atom_array.centre_atom_mask[
                    atom_array.mol_id == chain_id
                ].sum()
            chain_token_num = chain_ids_to_token_num[chain_id]

            if len(closest_chain_id) >= max_chains_num:
                if tokens + chain_token_num > max_tokens_num:
                    break

            closest_chain_id.add(chain_id)
            tokens += chain_token_num

        atom_array = atom_array[np.isin(atom_array.mol_id, list(closest_chain_id))]
        output_chains_num = len(np.unique(atom_array.mol_id))
        assert (
            output_chains_num == max_chains_num
            or atom_array.centre_atom_mask.sum() <= max_tokens_num
        )
        return atom_array, input_chains_num

    @staticmethod
    def remove_clashing_chains(
        atom_array: AtomArray,
        core_indices: list[int] = None,
    ) -> AtomArray:
        """
        Ref: AlphaFold3 SI Chapter 2.5.4

        Clashing chains are removed.
        Clashing chains are defined as those with >30% of atoms within 1.7 Å of an atom in another chain.
        If two chains are clashing with each other, the chain with the greater percentage of clashing atoms will be removed.
        If the same fraction of atoms are clashing, the chain with fewer total atoms is removed.
        If the chains have the same number of atoms, then the chain with the larger chain id is removed.

        Note: if two chains are covalent, they are not considered as clashing.

        Args:
            atom_array (AtomArray): Biotite AtomArray Object of a Bioassembly.
            core_indices (list[int]): A list of indices for core structures,
                                      where these indices correspond to structures that will be preferentially
                                      retained when pairwise clash chain assessments are performed.

        Returns:
            atom_array (AtomArray): An AtomArray that has been processed through this filter.
            removed_chain_ids (list[str]): A list of chain IDs that have been determined for deletion.
                                           This is to log whether the filter has been utilized.
        """
        chain_ids = np.unique(atom_array.chain_id[atom_array.is_resolved]).tolist()

        if core_indices is not None:
            core_chain_id = np.unique(atom_array.chain_id[core_indices])
        else:
            core_chain_id = np.array([])

        clash_records, chain_resolved_atom_nums = Filter._get_clashing_chains(
            atom_array, chain_ids
        )
        removed_chain_ids = Filter._get_removed_clash_chain_ids(
            clash_records,
            chain_ids,
            chain_resolved_atom_nums,
            core_chain_id=core_chain_id,
        )

        atom_array = atom_array[~np.isin(atom_array.chain_id, removed_chain_ids)]
        return atom_array, removed_chain_ids

    @staticmethod
    def remove_unresolved_mols(atom_array: AtomArray) -> AtomArray:
        """
        Remove molecules from a bioassembly object which all atoms are not resolved.

        Args:
            atom_array (AtomArray): Biotite AtomArray Object of a bioassembly.

        Returns:
            AtomArray: An AtomArray object with unresolved molecules removed.
        """
        valid_mol_id = []
        for mol_id in np.unique(atom_array.mol_id):
            resolved = atom_array.is_resolved[atom_array.mol_id == mol_id]
            if np.any(resolved):
                valid_mol_id.append(mol_id)

        atom_array = atom_array[np.isin(atom_array.mol_id, valid_mol_id)]
        return atom_array

    @staticmethod
    def remove_asymmetric_polymer_ligand_bonds(
        atom_array: AtomArray, entity_poly_type: dict
    ) -> AtomArray:
        """remove asymmetric polymer ligand bonds (including protein-protein bond, like disulfide bond).

        AF3 SI 5.1 Structure filters
        Bonds for structures with homomeric subcomplexes lacking the corresponding homomeric symmetry are also removed
        - e.g. if a certain bonded ligand only exists for some of the symmetric copies, but not for all,
        we remove the corresponding bond information from the input.
        In consequence the model has to learn to infer these bonds by itself.

        Args:
            atom_array (AtomArray): input atom array

        Returns:
            AtomArray: output atom array with asymmetric polymer ligand bonds removed.
        """
        # get inter chain bonds
        inter_chain_bonds = set()
        for i, j, b in atom_array.bonds.as_array():
            if atom_array.chain_id[i] != atom_array.chain_id[j]:
                inter_chain_bonds.add((i, j))

        # get asymmetric polymer ligand bonds
        asymmetric_bonds = set()
        chain_starts = struc.get_chain_starts(atom_array, add_exclusive_stop=False)
        for bond in inter_chain_bonds:

            if bond in asymmetric_bonds:
                continue

            i, j = bond
            atom_i = atom_array[i]
            atom_j = atom_array[j]
            i_is_polymer = atom_i.label_entity_id in entity_poly_type
            j_is_polymer = atom_j.label_entity_id in entity_poly_type
            if i_is_polymer:
                pass
            elif j_is_polymer:
                i, j = j, i
                atom_i, atom_j = atom_j, atom_i
                i_is_polymer, j_is_polymer = j_is_polymer, i_is_polymer
            else:
                # both entity is not polymer
                continue

            # get atom i mask from all entity i copies
            entity_mask_i = atom_array.label_entity_id == atom_i.label_entity_id
            num_copies = np.isin(chain_starts, np.flatnonzero(entity_mask_i)).sum()
            mask_i = (
                entity_mask_i
                & (atom_array.res_id == atom_i.res_id)
                & (atom_array.atom_name == atom_i.atom_name)
            )
            indices_i = np.flatnonzero(mask_i)

            if len(indices_i) != num_copies:
                # not every copy of entity i has atom i.
                asymmetric_bonds.add(bond)
                continue

            # check all atom i in entity i bond to an atom j in entity j.
            target_bonds = []
            for ii in indices_i:
                ii_bonds = [b for b in inter_chain_bonds if ii in b]
                for bond in ii_bonds:
                    jj = bond[1] if ii == bond[0] else bond[0]
                    atom_jj = atom_array[jj]
                    if atom_jj.label_entity_id != atom_j.label_entity_id:
                        continue
                    if atom_jj.res_name != atom_j.res_name:
                        continue
                    if atom_jj.atom_name != atom_j.atom_name:
                        continue
                    if j_is_polymer and atom_jj.res_id != atom_j.res_id:
                        # only for polymer, check res_id
                        continue
                    # found bond (ii, jj) with same enity_id, res_name, atom_name to bond (i,j)
                    target_bonds.append((min(ii, jj), max(ii, jj)))
                    break
            if len(target_bonds) != num_copies:
                asymmetric_bonds |= set(target_bonds)

        for bond in asymmetric_bonds:
            atom_array.bonds.remove_bond(bond[0], bond[1])
        return atom_array



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import uuid
from typing import Any, Sequence, Tuple

from protenix.utils.logger import get_logger
from protenix.web_service.colab_request_parser import RequestParser

logger = get_logger(__name__)


def need_msa_search(json_data: dict) -> bool:
    """
    Check if the input JSON data needs an MSA search.

    Args:
        json_data (dict): The input JSON data for a task.

    Returns:
        bool: True if an MSA search is required, False otherwise.
    """
    need_msa = False
    # the new format of msa filed is `pairedMsaPath` and `unpairedMsaPath`
    # we need to check `pairedMsaPath` and `unpairedMsaPath`
    for sequence in json_data["sequences"]:
        if "proteinChain" in sequence:
            protein_chain = sequence["proteinChain"]
            paired_msa_path = protein_chain.get("pairedMsaPath")
            unpaired_msa_path = protein_chain.get("unpairedMsaPath")

            if paired_msa_path is None and unpaired_msa_path is None:
                need_msa = True
            else:
                if paired_msa_path is not None and not os.path.exists(paired_msa_path):
                    logger.warning(
                        f"pairedMsaPath {paired_msa_path} does not exist, will re-search MSA."
                    )
                    need_msa = True
                if unpaired_msa_path is not None and not os.path.exists(
                    unpaired_msa_path
                ):
                    logger.warning(
                        f"unpairedMsaPath {unpaired_msa_path} does not exist, will re-search MSA."
                    )
                    need_msa = True
    return need_msa


def convert_msa_to_new_format(
    data_list: list[dict[str, Any]],
) -> tuple[list[dict[str, Any]], bool]:
    """
    Convert MSA format from old format to new format in a list of task dictionaries.

    Args:
        data_list (list[dict[str, Any]]): List of task dictionaries.

    Returns:
        tuple[list[dict[str, Any]], bool]:
            - Updated list of task dictionaries.
            - True if any format conversion occurred, False otherwise.
    """
    result = []
    json_need_converted = False
    for item in data_list:
        # Process each dictionary item
        processed_item, format_converted = convert_one_json_dict(item)
        json_need_converted = json_need_converted or format_converted
        result.append(processed_item)

    return result, json_need_converted


def convert_one_json_dict(obj: dict[str, Any]) -> tuple[dict[str, Any], bool]:
    """
    Process a single dictionary to convert 'msa' field to 'pairedMsaPath' and 'unpairedMsaPath'.

    Args:
        obj (dict[str, Any]): A single task dictionary.

    Returns:
        tuple[dict[str, Any], bool]: Updated dictionary and format conversion flag.
    """
    format_converted = False
    if "sequences" in obj and isinstance(obj["sequences"], list):
        for sequence in obj["sequences"]:
            if "proteinChain" in sequence and isinstance(
                sequence["proteinChain"], dict
            ):
                protein_chain = sequence["proteinChain"]
                if "msa" in protein_chain:
                    msa_info = protein_chain.pop("msa")  # Remove old msa field
                    logger.info(
                        f"Detecting old MSA format: {msa_info}, converting to new format."
                    )
                    precomputed_msa_dir = msa_info.get("precomputed_msa_dir")
                    format_converted = True
                    if precomputed_msa_dir:
                        # Build new paths
                        pairing_path = f"{precomputed_msa_dir}/pairing.a3m"
                        non_pairing_path = f"{precomputed_msa_dir}/non_pairing.a3m"

                        # Add new fields if the files exist
                        if os.path.exists(pairing_path):
                            protein_chain["pairedMsaPath"] = pairing_path
                        if os.path.exists(non_pairing_path):
                            protein_chain["unpairedMsaPath"] = non_pairing_path

    return obj, format_converted


def msa_search(
    seqs: Sequence[str], msa_res_dir: str, mode: str = "protenix"
) -> Sequence[str]:
    """
    Perform MSA search using MMseqs2 and return the resulting subdirectories.

    Args:
        seqs (Sequence[str]): List of protein sequences.
        msa_res_dir (str): Directory to save MSA results.
        mode (str): MSA search mode ('protenix' or 'colabfold').

    Returns:
        Sequence[str]: List of directories containing MSA results for each sequence.
    """
    os.makedirs(msa_res_dir, exist_ok=True)
    tmp_fasta_fpath = os.path.join(msa_res_dir, f"tmp_{uuid.uuid4().hex}.fasta")
    msa_res_subdirs = RequestParser.msa_search(
        seqs_pending_msa=seqs,
        tmp_fasta_fpath=tmp_fasta_fpath,
        msa_res_dir=msa_res_dir,
        mode=mode,
    )
    if mode == "protenix":
        msa_res_subdirs = RequestParser.msa_postprocess(
            seqs_pending_msa=seqs,
            msa_res_dir=msa_res_dir,
        )
    return msa_res_subdirs


def update_seq_msa(infer_seq: dict, msa_res_dir: str, mode: str) -> dict:
    """
    Update the sequences in the inference dictionary with their respective MSA paths.

    Args:
        infer_seq (dict): The task data containing sequences.
        msa_res_dir (str): Directory where MSA results are stored.
        mode (str): MSA search mode.

    Returns:
        dict: The updated task data.
    """
    protein_seqs = []
    for sequence in infer_seq["sequences"]:
        if "proteinChain" in sequence.keys():
            protein_seqs.append(sequence["proteinChain"]["sequence"])
    if len(protein_seqs) > 0:
        protein_seqs = sorted(protein_seqs)
        msa_res_subdirs = msa_search(protein_seqs, msa_res_dir, mode=mode)

        assert len(msa_res_subdirs) == len(protein_seqs), "msa search failed"
        protein_msa_res = dict(zip(protein_seqs, msa_res_subdirs))
        for sequence in infer_seq["sequences"]:
            if "proteinChain" in sequence.keys():
                precomputed_msa_dir = protein_msa_res[
                    sequence["proteinChain"]["sequence"]
                ]
                if os.path.exists(f"{precomputed_msa_dir}/pairing.a3m"):
                    sequence["proteinChain"][
                        "pairedMsaPath"
                    ] = f"{precomputed_msa_dir}/pairing.a3m"
                if os.path.exists(f"{precomputed_msa_dir}/non_pairing.a3m"):
                    sequence["proteinChain"][
                        "unpairedMsaPath"
                    ] = f"{precomputed_msa_dir}/non_pairing.a3m"

    return infer_seq


def update_infer_json(
    json_file: str, out_dir: str, use_msa: bool = True, mode: str = "protenix"
) -> Tuple[str, bool]:
    """
    Update the inference JSON file with MSA information.
    Iterates through tasks and runs MSA search if required and enabled.

    Args:
        json_file (str): Path to the input JSON file.
        out_dir (str): Directory to save MSA results.
        use_msa (bool): Whether to perform MSA search if missing.
        mode (str): MSA search mode ('protenix' or 'colabfold').

    Returns:
        Tuple[str, bool]:
            - Path to the updated (or original) JSON file.
            - Boolean indicating if any MSA search was actually performed.
    """
    if not os.path.exists(json_file):
        raise FileNotFoundError(f"Input file `{json_file}` does not exist.")
    with open(json_file, "r") as f:
        json_data = json.load(f)

    # Change the old format of msa filed to new format
    json_data, json_need_converted = convert_msa_to_new_format(json_data)

    actual_updated = False
    for task_idx, infer_data in enumerate(json_data):
        if use_msa and need_msa_search(infer_data):
            actual_updated = True
            task_name = infer_data.get("name", f"task_{task_idx}")
            logger.info(
                f"starting to update msa result for task {task_idx} in {json_file}"
            )
            update_seq_msa(
                infer_data,
                os.path.join(out_dir, task_name, "msa"),
                mode,
            )
    if actual_updated or json_need_converted:
        updated_json = os.path.join(
            os.path.dirname(os.path.abspath(json_file)),
            f"{os.path.splitext(os.path.basename(json_file))[0]}-update-msa.json",
        )
        with open(updated_json, "w") as f:
            json.dump(json_data, f, indent=4)
        logger.info(f"update msa result success and save to {updated_json}")
        return updated_json, actual_updated
    elif not use_msa:
        logger.warning(
            f"the inference json file {json_file} \n"
            "do not contain msa and will not be updated,\n"
            "and you set not using msa, in this mode, \n"
            "if you do not use esm feature, model performance "
            "might degrade significantly"
        )
        return json_file, actual_updated
    else:
        logger.info(f"do not need to update msa result, so return itself {json_file}")
        return json_file, actual_updated



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gzip
import json
import pickle
import time
from typing import Any, TYPE_CHECKING, Union

# Only import lmdb for type checking (no runtime cost)
if TYPE_CHECKING:
    import lmdb

import hashlib
import os
from pathlib import Path

import pandas as pd
import torch
from biotite import structure
from biotite.structure.io import pdbx

from protenix.data.constants import mmcif_restype_1to3
from protenix.utils.logger import get_logger
from protenix.utils.torch_utils import map_values_to_list, to_device

logger = get_logger(__name__)

PANDAS_NA_VALUES = [
    "",
    "#N/A",
    "#N/A N/A",
    "#NA",
    "-1.#IND",
    "-1.#QNAN",
    "-NaN",
    "-nan",
    "1.#IND",
    "1.#QNAN",
    "<NA>",
    "N/A",
    # "NA",
    "NULL",
    "NaN",
    "n/a",
    "nan",
    "null",
]

_INDICES_CSV_CACHE: dict[str, pd.DataFrame] = {}
_JSON_FILE_CACHE: dict[str, Any] = {}


def read_indices_csv(
    csv: Union[str, Path], load_indice_mode: str = "full"
) -> pd.DataFrame:
    """
    Read a csv file without the content changing.

    Args:
        csv (Union[str, Path]): A csv file path.
        load_indice_mode (str): Mode to load indices. Defaults to "full".

    Returns:
        pd.DataFrame : A pandas DataFrame.
    """
    csv_str = str(csv)

    if csv_str in _INDICES_CSV_CACHE:
        df_cached = _INDICES_CSV_CACHE[csv_str]
        logger.info(
            "[Protenix IO] read_indices_csv cache hit: path=%s, shape=%s",
            csv_str,
            tuple(df_cached.shape),
        )
        # Always return a copy so per-dataset filtering does not mutate the cache.
        return df_cached.copy(deep=True)

    t0 = time.time()
    if csv_str.endswith(".parquet"):
        if load_indice_mode == "simple":
            df = pd.read_parquet(
                csv_str,
                columns=["pdb_id", "type", "chain_1_id", "chain_2_id", "cluster_id"],
            )
        else:
            df = pd.read_parquet(csv_str)
        df = df.astype(str)
    else:
        df = pd.read_csv(
            csv_str, na_values=PANDAS_NA_VALUES, keep_default_na=False, dtype=str
        )

    def _normalize_entity_id(entity_id):
        if pd.isna(entity_id):
            return None
        try:
            return str(int(float(entity_id)))
        except (ValueError, TypeError):
            return None

    if "entity_1_id" in df.columns:
        df["entity_1_id"] = df["entity_1_id"].apply(_normalize_entity_id)
    if "entity_2_id" in df.columns:
        df["entity_2_id"] = df["entity_2_id"].apply(_normalize_entity_id)

    _INDICES_CSV_CACHE[csv_str] = df
    logger.info(
        "[Protenix IO] read_indices_csv finished in %.3fs: path=%s, shape=%s",
        time.time() - t0,
        csv_str,
        tuple(df.shape),
    )
    # Return a copy so callers can safely modify.
    return df.copy(deep=True)


class LMDBDict:
    """
    A lightweight dict-like wrapper around an LMDB database using SHA1-hashed keys.
    Values are read lazily from disk on access.

    Only record path and configuration at init time; do not open the
    Environment yet. Each process will open its own read-only env on the
    first access (safe under fork-based multiprocessing).

    Args:
        lmdb_path: Path to the LMDB database.
        lock (bool, optional): Whether to use locking. Defaults to False.
    """

    def __init__(self, lmdb_path, lock=False):
        self.path = str(lmdb_path)
        self.lock = lock
        # Per-process cache: pid -> lmdb.Environment
        self._env_by_pid: dict[int, Any] = {}  # Avoid runtime dependency on lmdb.Env

    def _get_env(self) -> "lmdb.Environment":
        """
        Maintain one read-only env per process to avoid sharing an Environment
        across forked processes.
        """
        pid = os.getpid()
        env = self._env_by_pid.get(pid)
        if env is None:
            # Lazy import: only import lmdb when actually needed
            try:
                import lmdb
            except ImportError as e:
                raise ImportError(
                    "LMDB is required to use LMDBDict. Please install it via: pip install lmdb"
                ) from e

            env = lmdb.open(
                self.path,
                subdir=False,
                readonly=True,
                lock=self.lock,
                readahead=False,
                meminit=False,
            )
            self._env_by_pid[pid] = env
        return env

    def _hash_key(self, key):
        """Compute SHA1 hash for a key, consistent with the writer script."""
        return hashlib.sha1(str(key).encode("utf-8")).hexdigest().encode("utf-8")

    def __getitem__(self, key):
        hashed_key = self._hash_key(key)
        env = self._get_env()
        with env.begin() as txn:
            value_bytes = txn.get(hashed_key)
            if value_bytes is None:
                raise KeyError(f"Key {key} not found in LMDB")
            # Assume the stored value is a UTF-8 string; fall back to raw bytes
            # to support arbitrary binary payloads.
            try:
                return value_bytes.decode("utf-8")
            except UnicodeDecodeError:
                return value_bytes

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def __contains__(self, key):
        hashed_key = self._hash_key(key)
        env = self._get_env()
        with env.begin() as txn:
            return txn.get(hashed_key) is not None

    def __len__(self):
        env = self._get_env()
        with env.begin() as txn:
            return txn.stat()["entries"]

    def keys(self):
        # Warning: do not convert keys() of a huge LMDB into a list().
        env = self._get_env()
        with env.begin() as txn:
            cursor = txn.cursor()
            for k, _ in cursor:
                # Note: yields hashed keys, original keys cannot be recovered here.
                yield k

    def close(self):
        for env in self._env_by_pid.values():
            try:
                env.close()
            except Exception:
                pass
        self._env_by_pid.clear()


def load_json_cached(path: Union[str, Path]) -> Any:
    """
    Load a JSON/Parquet/LMDB file with a simple in-process cache.

    - For JSON: Loads entire file into RAM (returns dict).
    - For Parquet: Loads into DataFrame (returns DataFrame).
    - For LMDB: Opens connection handle (returns LMDBDict wrapper).
    """
    path_str = str(Path(path))

    # 1. Cache JSON results per process; LMDB always returns a fresh wrapper.
    if path_str.endswith(".json") and path_str in _JSON_FILE_CACHE:
        logger.info("[Protenix IO] load_json_cached cache hit: path=%s", path_str)
        return _JSON_FILE_CACHE[path_str]

    t0 = time.time()

    # 2. Load according to suffix
    if path_str.endswith(".lmdb"):
        # LMDB branch: construct a new LMDBDict instance (not cached in _JSON_FILE_CACHE).
        data = LMDBDict(path_str)

    elif path_str.endswith(".json"):
        # JSON branch: load full JSON into memory and cache it.
        with open(path_str, "r") as f:
            try:
                import orjson

                data = orjson.loads(f.read())
            except ImportError:
                data = json.load(f)
        _JSON_FILE_CACHE[path_str] = data
    elif path_str == ".":
        # Empty JSON file
        data = {}
    else:
        raise ValueError(f"Unsupported file format: {path_str}")

    logger.info(
        "[Protenix IO] load_json_cached finished in %.3fs: path=%s",
        time.time() - t0,
        path_str,
    )
    return data


def load_gzip_pickle(pkl: Union[str, Path]) -> Any:
    """
    Load a gzip pickle file.

    Args:
        pkl (Union[str, Path]): A gzip pickle file path.

    Returns:
        Any: The loaded data.
    """
    with gzip.open(pkl, "rb") as f:
        data = pickle.load(f)
    return data


def dump_gzip_pickle(data: Any, pkl: Union[str, Path]):
    """
    Dump a gzip pickle file.

    Args:
        data (Any): The data to be dumped.
        pkl (Union[str, Path]): A gzip pickle file path.
    """
    with gzip.open(pkl, "wb") as f:
        pickle.dump(data, f)


def pdb_to_cif(pdb_file: str, cif_file: str, sequence=None):
    """
    Convert monomer protein pdb to CIF, and fill "entity_poly" and "entity_poly_seq"category.

    pdb_file (str): PDB file path.
    cif_file (str): output CIF file path.
    """
    entity_poly = pdbx.CIFCategory({"entity_id": "0", "type": "polypeptide(L)"})
    block_dict = {"entity_poly": entity_poly}

    if sequence is not None:
        entity_id = ["0"] * len(sequence)
        num = []
        mon_id = []
        for idx, i in enumerate(sequence):
            num.append(str(idx + 1))
            mon_id.append(mmcif_restype_1to3.get(i, "UNK"))
            assert (
                i in mmcif_restype_1to3 or i == "X"
            ), f"{i} is not in mmcif_restype_1to3"

        block_dict["entity_poly_seq"] = pdbx.CIFCategory(
            {
                "entity_id": entity_id,
                "num": num,
                "mon_id": mon_id,
            }
        )

    block = pdbx.CIFBlock(block_dict)
    pdb_file = Path(pdb_file)
    block_name = pdb_file.stem
    cif = pdbx.CIFFile({block_name: block})
    atom_array = structure.io.load_structure(pdb_file, extra_fields=["b_factor"])
    pdbx.set_structure(cif, atom_array)
    cif.write(cif_file)


class FloatEncoder(json.JSONEncoder):
    def __init__(self, *args, **kwargs):
        self.precision = kwargs.pop("precision", 2)
        super(FloatEncoder, self).__init__(*args, **kwargs)

    def encode(self, obj):
        def float_converter(o):
            if isinstance(o, float):
                return format(o, f".{self.precision}f")
            if isinstance(o, list):
                return [float_converter(i) for i in o]
            if isinstance(o, dict):
                return {k: float_converter(v) for k, v in o.items()}
            return o

        return super(FloatEncoder, self).encode(float_converter(obj))


def save_json(data, output_fpath, indent=4):
    data_json = data.copy()
    data_json = map_values_to_list(data_json)
    with open(output_fpath, "w") as f:
        if indent is not None:
            json.dump(data_json, f, indent=indent)
        else:
            json.dump(data_json, f)


def save_tensor(data, output_fpath):
    torch.save(to_device(data, device=None), output_fpath)



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from collections import defaultdict

import numpy as np
from rdkit import Chem


def neutralize_atoms(mol: Chem.Mol):
    pattern = Chem.MolFromSmarts(
        "[+1!h0!$([*]~[-1,-2,-3,-4]),-1!#4!#5!$([*]~[+1,+2,+3,+4])]"
    )
    at_matches = mol.GetSubstructMatches(pattern)
    at_matches_list = [y[0] for y in at_matches]
    if len(at_matches_list) > 0:
        for at_idx in at_matches_list:
            atom = mol.GetAtomWithIdx(at_idx)
            chg = atom.GetFormalCharge()
            hcount = atom.GetTotalNumHs()
            atom.SetFormalCharge(0)
            atom.SetNumExplicitHs(hcount - chg)
            atom.UpdatePropertyCache()
    return mol


def recursive_permutation(atom_inds, permutation_list, res):
    def _permute_atom_ind(atom_inds, permutation):
        # atom_inds: list of atom (positional) indices
        # permutation: values to be permutated in the given order
        permute_inds = [i for i, a in enumerate(atom_inds) if a in permutation]
        for i, perm_ind in enumerate(permute_inds):
            atom_inds[perm_ind] = permutation[i]
        return atom_inds

    if len(permutation_list) == 0:
        res.append(atom_inds)
    else:
        current_permutation_list = permutation_list.copy()
        for permutation in current_permutation_list.pop(0):
            atom_inds_permed = _permute_atom_ind(atom_inds.copy(), permutation)
            recursive_permutation(atom_inds_permed, current_permutation_list, res)


def augment_atom_maps_with_conjugate_terminal_groups(
    original_maps, atomic_number_mapping, terminal_group_tuples, MaxMatches=1e6
):
    """
    Augment atom maps from GetSubstructMatches with extra symmetry from confjugated terminal groups.
    Parameters
    --------------
    original_maps:  Tuple(Tuples), all possible atom index mappings, note we require that the mappings should range from 0 to n_heavy_atom-1 (a.k.a. no gap in indexing)
    atomic_number_mapping: dict, mapping from atom (positional) indices to its atomic numbers, for splitting/removing different types of atoms in each terminal group
    terminal_group_tuples: Tuple(Tuples), a group of pair of atoms whose bonds match the SMARTS string. Ex: ((0, 1), (2, 1), (10, 9), (11, 9), (12, 9), (14, 13), (15, 13))
    MaxMatches: int, cutoff for total number of matches (n_original_perm * n_conjugate perm)

    Returns
    --------------
    augmented_maps: Tuple(Tuples) , original_maps augmented by muliplying the permutations induced by terminal_group_tuples.
    """

    def _terminal_atom_cluster_from_pairs(edges):
        graph = defaultdict(set)
        for u, v in edges:
            graph[u].add(v)
            graph[v].add(u)
        return graph

    def _split_sets_by_mapped_values(list_of_sets, mapping):
        result = []
        for s in list_of_sets:
            mapped_sets = {}
            for elem in s:
                mapped_value = mapping.get(elem)
                if mapped_value not in mapped_sets:
                    mapped_sets[mapped_value] = set()
                mapped_sets[mapped_value].add(elem)
            result.extend(mapped_sets.values())
        return result

    # group terminal group tuples with common atom_indices: [{0, 2}, {10, 11, 12}, {14, 15}]
    terminal_atom_clusters = _terminal_atom_cluster_from_pairs(terminal_group_tuples)
    MaxTerminalGroups = max(
        1, int(np.ceil(np.emath.logn(3, MaxMatches / len(original_maps))))
    )
    # if MaxTerminalGroups is less than the total number terminal groups, sample the first {MaxTerminalGroups} groups (to remove randomness)

    perm_groups = sorted(
        [
            atom_inds
            for common_id, atom_inds in terminal_atom_clusters.items()
            if len(atom_inds) > 1
        ]
    )[: min(MaxTerminalGroups, len(terminal_atom_clusters))]

    # within each terminal group, if there are different atom types, split by atom type (if only one left, discard)
    perm_groups = _split_sets_by_mapped_values(perm_groups, atomic_number_mapping)
    perm_groups = [p for p in perm_groups if len(p) > 1]

    # all permutations according to symmetric conjugate terminal atoms: [[(0, 2), (2, 0)], [(10, 11, 12), (10, 12, 11), (11, 10, 12), (11, 12, 10), (12, 10, 11), (12, 11, 10)], [(14, 15), (15, 14)]]
    perm_groups = [sorted(list(itertools.permutations(g))) for g in perm_groups]

    # recursively permute the original mappings
    augmented_maps = []
    for initial_mapping in original_maps:
        recursive_permutation(list(initial_mapping), perm_groups, augmented_maps)

    # Convert to the same data type as in original_maps
    augmented_maps = tuple(tuple(a) for a in augmented_maps)
    # Remove duplicates: original_maps might have already permutated some of the conjugate_terminal group indices
    return tuple(set(augmented_maps))


def _get_substructure_perms(
    mol: Chem.Mol,
    Neutralize: bool = False,
    CheckStereochem: bool = True,
    SymmetrizeConjugatedTerminal: bool = True,
    MaxMatches: int = 512,
) -> np.ndarray:
    """
    Args:
        CheckStereochem: whether to assure stereochem does not change after permutation
        Neutralize: if true, neutralize the mol before computing the permutations
        SymmetrizeConjugatedTerminal: if true, consider symmetrization of conjugated terminal groups
        MaxMatches: int, cutoff for total number of matches

    return shape=[num_perms, num_atoms]
    """
    ori_idx_w_h = []
    for atom in mol.GetAtoms():
        atom.SetProp("ori_idx_w_h", str(atom.GetIdx()))
        ori_idx_w_h.append(atom.GetIdx())

    # Attention !!!
    # Remove Hs; Otherwise, there will be too many matches.
    mol = Chem.RemoveHs(mol)
    if Neutralize:
        mol = neutralize_atoms(mol)

    # Get substructure matches
    base_perms = np.array(
        mol.GetSubstructMatches(mol, uniquify=False, maxMatches=MaxMatches)
    )
    assert len(base_perms) > 0, "no matches found, error"
    # Check stereochem
    if CheckStereochem:
        chem_order = np.array(
            list(Chem.rdmolfiles.CanonicalRankAtoms(mol, breakTies=False))
        )
        perms_mask = (chem_order[base_perms] == chem_order[None]).sum(
            -1
        ) == mol.GetNumAtoms()
        base_perms = base_perms[perms_mask]

    # Add terminal conjugate groups
    sma = "[O,N;D1;$([O,N;D1]-[*]=[O,N;D1]),$([O,N;D1]=[*]-[O,N;D1])]~[*]"
    patt = Chem.MolFromSmarts(sma)
    terminal_group_tuples = mol.GetSubstructMatches(patt)
    if (
        len(terminal_group_tuples) > 0 and SymmetrizeConjugatedTerminal
    ):  # Only augment if there exist conjugate pairs or if user sets to
        atomic_number_mapping = {
            i: atom.GetAtomicNum() for i, atom in enumerate(mol.GetAtoms())
        }
        base_perms = augment_atom_maps_with_conjugate_terminal_groups(
            tuple(tuple(a) for a in base_perms),
            atomic_number_mapping,
            terminal_group_tuples,
            MaxMatches,
        )
        base_perms = np.array(base_perms)

    if len(base_perms) > MaxMatches:
        base_perms = base_perms[:MaxMatches]

    new_to_ori_idx_map = {}
    ori_to_new_idx_map = {}
    for atom in mol.GetAtoms():
        ori_idx = int(atom.GetProp("ori_idx_w_h"))
        new_idx = atom.GetIdx()
        new_to_ori_idx_map[new_idx] = ori_idx
        ori_to_new_idx_map[ori_idx] = new_idx

    base_perms = np.vectorize(new_to_ori_idx_map.get)(base_perms)
    perms = np.zeros(shape=(base_perms.shape[0], len(ori_idx_w_h)))
    for i in range(len(ori_idx_w_h)):
        if i in ori_to_new_idx_map:
            perms[:, i] = base_perms[:, ori_to_new_idx_map[i]]
        else:
            # The position of the H atom will not be exchanged.
            perms[:, i] = i
    return perms


def get_substructure_perms(
    mol: Chem.Mol,
    CheckStereochem: bool = True,
    SymmetrizeConjugatedTerminal: bool = True,
    MaxMatches: int = 512,
    KeepProtonation: bool = False,
) -> np.ndarray:
    kwargs = {
        "CheckStereochem": CheckStereochem,
        "SymmetrizeConjugatedTerminal": SymmetrizeConjugatedTerminal,
        "MaxMatches": MaxMatches,
    }

    if KeepProtonation:
        perms = _get_substructure_perms(mol, Neutralize=False, **kwargs)
    else:
        # Have to deuplicate permutations across the two protonation states
        perms = np.unique(
            np.vstack(
                (
                    _get_substructure_perms(mol, Neutralize=False, **kwargs),
                    _get_substructure_perms(mol, Neutralize=True, **kwargs),
                )
            ),
            axis=0,
        )

    nperm = len(perms)
    if nperm > MaxMatches:
        perms = perms[np.random.choice(range(nperm), MaxMatches, replace=False)]
    return perms


def test():
    testcases = [
        "C1=CC=CC=C1",
        "CC(=O)OC1=CC=CC=C1C(=O)O",
        "C[C@H](CCC(=O)O)[C@H]1CC[C@@H]2[C@@]1(CC[C@H]3[C@H]2CC=C4[C@@]3(CC[C@@H](C4)O)C)C",
        "CN1C=NC2=C1C(=O)N(C(=O)N2C)C",
    ]
    for smiles in testcases:
        print(smiles)
        molecule = Chem.MolFromSmiles(smiles)
        perms = get_substructure_perms(molecule)
        print(perms.shape)
        print(perms.T)


if __name__ == "__main__":
    test()



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import traceback

import pandas as pd
import torch

from protenix.data.esm.compute_esm import compute_ESM_embeddings, load_esm_model
from protenix.utils.logger import get_logger

logger = get_logger(__name__)


class ESMFeaturizer:
    def __init__(
        self,
        embedding_dir: str,
        sequence_fpath: str,
        embedding_dim: int = 1028,
        error_dir: str = None,
    ):
        self.embedding_dir = embedding_dir
        self.sequence_fpath = sequence_fpath
        self.seq_to_filename = self.get_seq_to_filename(sequence_fpath)
        self.embedding_dim = embedding_dim
        self.error_dir = error_dir
        if self.error_dir is not None:
            self.error_dir = os.path.join(self.error_dir, "esm_error")
            os.makedirs(self.error_dir, exist_ok=True)

    def get_seq_to_filename(self, sequence_fpath: str) -> dict[str, str]:
        df = pd.read_csv(sequence_fpath)
        df["filename"] = (
            df["part_id"].astype(str) + "/" + df["seq_label"].astype(str) + ".pt"
        )
        return df.set_index("seq")["filename"].to_dict()

    def load_esm_embedding(self, sequence: str):
        x = torch.load(
            os.path.join(self.embedding_dir, self.seq_to_filename[sequence]),
            weights_only=True,
        )
        assert x.size(0) == len(sequence), (
            f"ESM embedding size {x.size(0)} not equal to sequence length {len(sequence)}"
            + "The error occurs because embeddings were previously computed for a certain 'name', "
            + "and then a different JSON file with the same task name was used, causing ESM to fail loading the existing embeddings."
            + "You can resolve this by deleting the local esm_embeddings directory and retrying."
            + "We recommend that the 'name' field in the JSON file be unique."
        )
        return x

    def save_error(self, error_sequences, pdb_id):
        if (self.error_dir is None) or (len(error_sequences) == 0):
            return
        for error_data in error_sequences:
            fpath = os.path.join(
                self.error_dir, f"{pdb_id}_{error_data['entity_id']}.txt"
            )
            if os.path.exists(fpath):
                continue
            with open(fpath, "w") as f:
                f.write(error_data["error"])

    def __call__(self, token_array, atom_array, bioassembly_dict, inference_mode=False):

        # init as zeros
        N_token = len(token_array)
        x = torch.zeros([N_token, self.embedding_dim])

        # get one atom per token
        centre_atoms_indices = token_array.get_annotation("centre_atom_index")
        centre_atom_array = atom_array[centre_atoms_indices]

        if inference_mode:
            entity_id_to_sequence = (
                {}
            )  # Only contains protein entity, many-to-one mapping
            for i, entity_info_wrapper in enumerate(bioassembly_dict["sequences"]):
                entity_id = str(i + 1)
                entity_type = list(entity_info_wrapper.keys())[0]
                entity_info = entity_info_wrapper[entity_type]
                if entity_type == "proteinChain":
                    entity_id_to_sequence[entity_id] = entity_info["sequence"]
            protein_entity_ids = set(entity_id_to_sequence.keys())
        else:
            is_protein = centre_atom_array.chain_mol_type == "protein"
            protein_entity_ids = set(centre_atom_array.label_entity_id[is_protein])

        # enumerate over the entities
        error_sequences = []
        for entity_id in protein_entity_ids:
            try:
                # Get sequence
                if inference_mode:
                    sequence = entity_id_to_sequence[entity_id]
                else:
                    sequence = bioassembly_dict["sequences"][str(entity_id)]
                x_esm = self.load_esm_embedding(sequence)
                # Get residue indices of the cropped tokens
                entity_mask = centre_atom_array.label_entity_id == entity_id
                res_index = (
                    centre_atom_array.res_id[entity_mask] - 1
                )  # res_id starts with 1
                # Get esm embeddding according to residue indices
                x[entity_mask] = x_esm[res_index]
            except Exception as e:
                error_message = f"{e}:\n{traceback.format_exc()}"
                error_sequences.append(
                    {
                        "entity_id": entity_id,
                        "error": error_message,
                    }
                )
                logger.warning(
                    f"[{bioassembly_dict['pdb_id']}] ESM error: {error_message}"
                )

        id_key = "name" if inference_mode else "pdb_id"
        self.save_error(error_sequences, pdb_id=bioassembly_dict[id_key])

        return x

    @staticmethod
    def precompute_esm_embedding(
        inputs: list, model_name, embedding_dir, sequence_fpath, checkpoint_dir
    ):
        print("Precompute ESM embeddings")
        # prepare seq_label
        all_seq_dict = []
        for sample_dict in inputs:
            sample_name = sample_dict["name"]
            for i, entity_info_wrapper in enumerate(sample_dict["sequences"]):
                pdb_entity_id = sample_name + "_" + str(i + 1)
                entity_type = list(entity_info_wrapper.keys())[0]
                entity_info = entity_info_wrapper[entity_type]
                if entity_type == "proteinChain":
                    all_seq_dict.append(
                        {
                            "seq": entity_info["sequence"],
                            "pdb_entity_id": pdb_entity_id,
                            "seq_label": pdb_entity_id,
                            "part_id": pdb_entity_id,
                        }
                    )
        df_seq = pd.DataFrame(
            all_seq_dict, columns=["seq", "pdb_entity_id", "seq_label", "part_id"]
        )
        df_seq.to_csv(sequence_fpath)
        print(f"Save sequence file to {sequence_fpath}")

        model, alphabet = load_esm_model(model_name, local_esm_dir=checkpoint_dir)
        error_parts = []
        part_counts = dict(df_seq["part_id"].value_counts())
        for part_id, count in part_counts.items():
            df_part = df_seq[df_seq["part_id"] == part_id]
            print(f"Part {part_id}: {len(df_part)} sequences.")
            labels = df_part["seq_label"].tolist()
            sequences = df_part["seq"].tolist()
            try:
                save_dir = os.path.join(embedding_dir, part_id)
                if not os.path.exists(save_dir):
                    os.makedirs(save_dir)
                lm_embeddings = compute_ESM_embeddings(
                    model_name,
                    model,
                    alphabet,
                    labels,
                    sequences,
                    save_dir,
                    truncation_seq_length=4094,
                    toks_per_batch=16384,
                )
                print(
                    f"[{part_id}] Processed {len(lm_embeddings)} sequences in total. Done!"
                )
            except Exception as e:
                print(f"[{part_id}] {e}")
                error_parts.append(part_id)
        print("Error parts: ", error_parts)



# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import contextlib
import shutil
import tempfile
from typing import Any, Iterable, List, Optional, Sequence, Tuple


@contextlib.contextmanager
def tmpdir_manager(base_dir: Optional[str] = None):
    """Context manager that deletes a temporary directory on exit."""
    tmpdir = tempfile.mkdtemp(dir=base_dir)
    try:
        yield tmpdir
    finally:
        shutil.rmtree(tmpdir, ignore_errors=True)


def to_a3m(sequences: Sequence[str]) -> str:
    """Converts sequences to an a3m file."""
    names = [f"sequence {i}" for i in range(1, len(sequences) + 1)]
    a3m = []
    for sequence, name in zip(sequences, names):
        a3m.append(">" + name + "\n")
        a3m.append(sequence + "\n")
    return "".join(a3m)


def parse_fasta(data: str) -> Tuple[List[str], List[str]]:
    """
    Parse FASTA/A3M string into sequences and descriptions.

    Args:
        data: The input FASTA or A3M formatted string.

    Returns:
        A tuple of (sequences, descriptions).
    """
    sequences = []
    descriptions = []
    index = -1
    for line in data.splitlines():
        line = line.strip()
        if not line or line.startswith("#"):
            continue
        if line.startswith(">"):
            index += 1
            descriptions.append(line[1:])
            sequences.append("")
        else:
            if index >= 0:
                sequences[index] += line
    return sequences, descriptions


def parse_kalign_a3m(a3m_string: str) -> List[str]:
    """Parses sequences from an A3M format alignment."""
    sequences, _ = parse_fasta(a3m_string)
    return sequences


def lazy_fasta_parse(data: str) -> Iterable[Tuple[str, str]]:
    """
    Memory-friendly FASTA parsing yielding (sequence, description) pairs.

    Args:
        data: The input FASTA or A3M formatted string.

    Yields:
        Tuples of (sequence, description).
    """
    curr_desc, curr_seq = None, []
    for line in data.splitlines():
        line = line.strip()
        if not line:
            continue
        if line.startswith(">"):
            if curr_desc:
                yield "".join(curr_seq), curr_desc
            curr_desc, curr_seq = line[1:].strip(), []
        else:
            curr_seq.append(line)
    if curr_desc:
        yield "".join(curr_seq), curr_desc


def a3m_to_sto_list(a3m_seqs: List[str]) -> List[str]:
    """
    Convert a list of A3M sequences to Stockholm format aligned sequences.

    Args:
        a3m_seqs: List of sequences in A3M format.

    Returns:
        List of aligned sequences in Stockholm format.

    Raises:
        ValueError: If A3M rows have inconsistent lengths.
    """
    if not a3m_seqs:
        return []
    cursors = [list(s) for s in a3m_seqs]
    out = [""] * len(a3m_seqs)
    while any(cursors):
        if any(c and c[0].islower() for c in cursors):
            for i, c in enumerate(cursors):
                if c and c[0].islower():
                    out[i] += c.pop(0).upper()
                else:
                    out[i] += "-"
        else:
            for i, c in enumerate(cursors):
                if not c:
                    raise ValueError("Inconsistent row lengths in A3M")
                out[i] += c.pop(0)
    return out


def align_to_query(seq: str, query: str) -> str:
    """
    Align sequence to a gapless query sequence.

    Args:
        seq: Sequence to align.
        query: Reference query sequence.

    Returns:
        Aligned sequence string.

    Raises:
        ValueError: If lengths do not match.
    """
    if len(seq) != len(query):
        raise ValueError("Length mismatch")
    return "".join(
        s if q != "-" else (s.lower() if s != "-" else "") for s, q in zip(seq, query)
    )


def convert_a3m_to_stockholm(a3m: str, max_seqs: Optional[int] = None) -> str:
    """
    Convert A3M format string to Stockholm format string.

    Args:
        a3m: Input A3M format string.
        max_seqs: Maximum number of sequences to include.

    Returns:
        Stockholm format string.
    """
    seqs, descs = parse_fasta(a3m)
    if max_seqs:
        seqs, descs = seqs[:max_seqs], descs[:max_seqs]

    names = [f"{d.split()[0]}_{i}" for i, d in enumerate(descs)]
    gs_lines = [
        f"#=GS {n} DE {d.partition(' ')[2].strip() or '<EMPTY>'}"
        for n, d in zip(names, descs)
    ]
    aligned = a3m_to_sto_list(seqs)

    w = max(len(n) for n in names + ["#=GC RF"])
    msa_lines = [f"{n:<{w}} {s}" for n, s in zip(names, aligned)]
    rf = "".join("." if c == "-" else "x" for c in aligned[0])

    return "\n".join(
        ["# STOCKHOLM 1.0", ""]
        + gs_lines
        + [""]
        + msa_lines
        + [f'{"#=GC RF":<{w}} {rf}', "//"]
    )


def convert_stockholm_to_a3m(
    sto_io: Any,
    max_seqs: Optional[int] = None,
    remove_gaps: bool = True,
    linewidth: Optional[int] = None,
) -> str:
    """
    Convert Stockholm format from an IO object to A3M format string.

    Args:
        sto_io: Iterable of Stockholm lines.
        max_seqs: Maximum sequences to parse.
        remove_gaps: Whether to remove gaps in the first row.
        linewidth: Optional line width for the output sequence.

    Returns:
        A3M format string.
    """
    seqs, descs = {}, {}
    for line in sto_io:
        line = line.strip()
        if not line or line.startswith(("//")):
            continue
        if line.startswith("#=GS"):
            parts = line.split(maxsplit=3)
            if len(parts) >= 3 and parts[2] == "DE":
                descs[parts[1]] = parts[3] if len(parts) == 4 else ""
        elif not line.startswith("#"):
            parts = line.split(maxsplit=1)
            if len(parts) == 2:
                name, seq = parts
                if name not in seqs and max_seqs and len(seqs) >= max_seqs:
                    continue
                seqs[name] = seqs.get(name, "") + seq

    if not seqs:
        return ""
    query_seq = next(iter(seqs.values()))
    chunks = []
    for name, s in seqs.items():
        chunks.append(f">{name} {descs.get(name, '')}")
        final_s = (
            align_to_query(s, query_seq).replace(".", "")
            if remove_gaps
            else s.replace(".", "")
        )
        if linewidth:
            chunks.extend(
                final_s[i : i + linewidth] for i in range(0, len(final_s), linewidth)
            )
        else:
            chunks.append(final_s)
    return "\n".join(chunks) + "\n"
